#include <stdio.h>
#include <limits.h>
#include <math.h>  
#include <nicklib.h> 
#include <getpars.h> 

#include "admutils.h" 

char *iname = NULL ;
int verbose = NO ;
int xmode = NO ;  // YES just one sigma 

double *gamms, *qobs ;


void readcommands(int argc, char **argv) ;
int readit (char *iname) ;
int getsize (char *iname) ;

int main(int argc, char **argv)
{
    int iter, numiter = 100 ;   
    int n=10000 ; 
    int dim =3 ;
    int i, j, k ;
    long seed ; 
    double *mm, *sig, *ww, *gg, *s0, *s1, *s2 ;
    double z, y, ylike, ybase, ynull, yn ;
    double ychi ;
    double rmean, rsig ;
    int dof=6 ;

    readcommands(argc, argv) ;
    seed = 77 ; 
    SRAND(seed) ;

    n = getsize(iname) ;

    printf("## iname: %s  n: %d\n", iname, n) ;

    ZALLOC(gamms, n*dim, double) ;  
    ZALLOC(qobs,  n, double) ;  
    ZALLOC(mm,    dim, double) ;  
    ZALLOC(sig,   dim, double) ;  
    ZALLOC(ww,    dim, double) ;  

    readit(iname) ;

    yn = (double) n ;
    rmean = asum(qobs, n)/yn ;
    y = asum2(qobs, n) / yn ;
    y -= (rmean*rmean) ;
    rsig = sqrt(y) ; 
    printf("##rmean: %9.3f  rsig: %9.3f\n", rmean, rsig) ;

    mm[0] = 0 ;
    mm[1] = 1 ;
    mm[2] = 2 ;
    vclear(sig, 1.0, 3) ;

    vzero(mm, dim) ;
    vclear(sig, 2.0, dim) ;
    ZALLOC(gg, dim, double) ;
    ZALLOC(s0, dim, double) ;
    ZALLOC(s1, dim, double) ;
    ZALLOC(s2, dim, double) ;

    ynull = 0.0 ;
    for (iter = 1; iter <= numiter; ++iter) {  
     vzero(s0, dim) ;
     vzero(s1, dim) ;
     vzero(s2, dim) ;
     ylike = 0.0 ;
     for (i=0; i<n; i++) {  
      if (iter==1) {  
       z = qobs[i]-rmean ; z /= rsig ;
       y = nordis(z)/rsig ; 
       ynull += log(y) ;
      }
      for (j=0; j<dim; j++) {  
       z = qobs[i] - mm[j] ; z /= sig[j] ;
       gg[j] = nordis(z) * gamms[i*dim+j]/sig[j] ;
      }
      y = bal1(gg, dim) ;
      ylike += log(y) ;
      for (j=0; j<dim; j++) {  
       s0[j] += gg[j] ;
       s1[j] += qobs[i]*gg[j] ;
       s2[j] += qobs[i]*qobs[i]*gg[j] ;
      }
     }
      if (iter==1) ybase = ynull ;
      ylike -= ybase ;
      printf("%4d %9.3f\n", iter, ylike) ;
/**
      printf("zz %d \n", n) ;
      printmat(gg, 1, 3) ;
      printmat(s0, 1, 3) ;
      printmat(s1, 1, 3) ;
      printmat(s2, 1, 3) ;
*/
      vvd(mm, s1, s0, dim) ;
      vvd(s2, s2, s0, dim) ;
      vvt(ww, mm, mm, dim) ;
      vvm(s2, s2, ww, dim) ;   // variance 
      if (xmode) { 
        y = vdot(s0, s2, dim) / asum(s0, dim) ;
        vclear(s2, y, dim) ;
        dof = 4 ;
      }
      vsqrt(sig, s2, dim) ;
    }
    printf("mm: ") ; printmat(mm, 1, dim) ;
    printf("sig: ") ; printmat(sig, 1, dim) ;
    ychi = 2.0*ylike ;
    printf("LRT : %9.3f dof: %d\n", ychi, dof-2) ;
}

void readcommands(int argc, char **argv) 

{
  int i ;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "i:vVx")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'i':
	iname = strdup(optarg) ;
	break;

      case 'x':
	xmode = YES ;
	break; 


      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }
}

#define MAXSTR 512 
#define MAXFF  20

int readit(char *iname)
{

  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  double xx[4] ;
  int nsplit, i, j, num=0 ;
  FILE *fff ;

  openit(iname, &fff, "r") ;

  while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   for (j=0; j<4; ++j) {  
    xx[j] = atof(spt[j]) ;
   }
   copyarr(xx, gamms + num*3, 3) ;
   qobs[num] = xx[3] ;
   freeup(spt, nsplit) ;
   ++num ;
  }
  fclose(fff) ;
  return num ;
}

int getsize(char *iname)
{

  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  double xx[4] ;
  int nsplit, i, j, num=0 ;
  FILE *fff ;

  openit(iname, &fff, "r") ;

  while (fgets(line, MAXSTR, fff) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   freeup(spt, nsplit) ;
   ++num ;
  }
  fclose(fff) ;
  return num ;
}

