#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "badpairs.h"
#include "admutils.h"
#include "mcio.h"  
#include "mcmcpars.h"  
#include "eigsubs.h"  

#define WVERSION   "3000" 
/** 
Simple eigenvector analysis
Options to look at groups (simple ANOVA)
Weights allowed for individuals 
missing mode
dotpops added
recompiled with new twtail. Output form at changed
Cleaned up twestxx
fancynorm mode (divide by sqrt(p*(1-p)) 
poplistname supported.  Eigenanalysis just on individuals in population 
But all individuals figure in eigenvector output
New way of computing effective marker size  (twl2mode)
popdifference implemented
ldregression  ldlimit (genetic distance in Morgans)
nostatslim added  
dotpop has new format if many groups
uses new I/O
*/


#define MAXFL  50   
#define MAXSTR  512
#define MAXPOPS 100

char *parname = NULL ;
char *trashdir = "/var/tmp" ;
int verbose = NO ;
int qtmode = NO ;
int numperms  = 0  ; // permutation testing
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 
int numeigs = 5 ;
int markerscore = NO ;
int seed = 0 ;
int chisqmode = NO ;  // approx p-value better to use F-stat
int missingmode = NO ;
int plotmode = NO ;
int maxmode = NO ;  /* eigenvector notmalization */
int fancynorm = YES ;
int dotpopsmode = YES ;
int noxdata = YES ;  /* default as pop structure dubious if Males and females */
int pcorrmode = NO ;
int pcpopsonly = YES ;
int rhomode = NO ;
int rhonums  =  1 ;
int nostatslim = 10 ;

double plo = .001 ;
double phi = .999 ;
double pvhit = .001 ;
double pvjack = 1.0e-6 ;
double *chitot ;
int    *xpopsize ;

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *snpoutfilename = NULL ;
char  *indivname = NULL ;
char *badsnpname = NULL ;
char *poplistname = NULL ;

int ldregress = 0 ;
double ldlimit = 9999.0 ;  /* default is infinity */
/* we only consider markers as in possible LD if gdis <= ldlimit */

char *outputname = NULL ;
char *weightname = NULL ;
FILE *ofile ;

double twestxx(double *lam, int m, double *pzn,  double *pzvar)  ;
double twnorm(double lam, double m, double n) ;

void readcommands(int argc, char **argv) ;
int loadindx(Indiv **xindlist, int *xindex, Indiv **indivmarkers, int numindivs) ;
int loadsnpx(SNP **xsnplist, SNP **snpmarkers, int numsnps, Indiv **indivmarkers) ;
void loadxdata(double *xdata, SNP **snplist, int *xindex, int nrows, int ncols) ;           
void loadxdataind(double *xdata, SNP **snplist, int ind,  int ncols) ;           
void fixxrow(double *xrow, double *xmean, double *xfancy, int len)  ;
void coladjust(double *xdata, int nrows, int ncols, 
 int fancy, double *xm, double *xf, double *tauv)   ;
void dofancy(double *cc, int n, double *fancy) ;
int vadjust(double *rr, int n, double *pmean)  ;
void getcol(double *cc, double *xdata, int col, int nrows, int ncols)  ;
void putcol(double *cc, double *xdata, int col, int nrows, int ncols)  ;
double dottest(char *sss, double *vec, char **eglist, int numeg, int *xtypes, int len) ;
double yll(double x1, double x2, double xlen) ;
void calcmean(double *wmean, double *vec, int len, int *xtypes, int numeg)  ;
double anova1(double *vec, int len, int *xtypes, int numeg) ;
double anova(double *vec, int len, int *xtypes, int numeg) ;
void publishit(char *sss, int df, double chi) ;
void
permtest(int *numhits, double *evals, int numeigs, double *xdata, 
 int numperms, int nrows, int ncols, int weightmode, SNP **xsnplist) ;

void randperm(double *x, int n)  ;
void setmiss(SNP **snpm, int numsnps)  ;
void setfvecs(double *fvecs, double *evecs, int nrows, int numeigs) ;
void dotpops(double *X, char **eglist, int numeg, int *xtypes, int nrows)  ;
void printxcorr(double *X, int nrows, Indiv **indxx)  ;      

void ldreg(double *ldmat, double *ldmat2, 
  double *vv, double *vv2, double *ldvv, 
  double *ldvv2, int rsize, int n)  ;

void clearld(double *ldmat, double *ldvv, int rsize, int n, int nclear)  ;

void
setidmat(double *mat, int n, double scal)  ; 

void fixrho(double *a, int n) ;
void printdiag(double *a, int n) ;
void printdiagt(double *a, double *xtau, int n) ;

void
dotau(double *xdata, int nrows, int ncols, double *tauvec)  ;
double esttau(double *pp, double *cc, int n) ;
double easyesttau(double *pp, double *cc, int n) ;

int main(int argc, char **argv)
{

  char sss[MAXSTR] ;
  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  int  lsnplist, lindlist, numeg ;
  int i, j, k, k1, k2; 
  SNP *cupt, *cupt1, *cupt2, *cupt3 ;
  Indiv *indx ;
  double y1, y2, y ;

  int ch1, ch2 ;
  int fmnum , lmnum ;
  int num, n1, n2 ;

  int nindiv = 0, e, f, lag=1  ;
  double xc[9], xd[4], xc2[9] ;
  double ychi,  tail, tw ;
  int nignore, numrisks = 1 ;
  double *xdata, *xrow ; 
  SNP **xsnplist  ;
  Indiv **xindlist ;
  int *xindex, *xtypes ;
  int nrows, ncols, m ;
  double *XTX, *cc, *evecs, *ww, weight, *qvec, *qcoord ; 
  double *tvecs, *ttvecs ;
  double zn, zvar ;
  double *fvecs, *fxvecs, *fxscal ;
  double *ffvecs ;
  int weightmode = NO ;
  double chisq ;
  int *numhits, t ;  
  double *xmean, *xfancy ;
  double *ldmat, *ldmat2, *ldvv, *ldvv2, *vv2  ;
  int chrom, lastchrom, numclear ;
  double gdis ;
  double *tauvec = NULL ;

  readcommands(argc, argv) ;
  printf("## eigind version: %s\n", WVERSION) ;
  if (parname == NULL) return 0 ;

  if (fancynorm) printf("fancynorm used\n") ;
  else printf("no fancynorm used\n") ;

  nostatslim = MAX(nostatslim, 3) ;

  ofile = stdout; 
  if (outputname != NULL) openit(outputname, &ofile, "w") ;

  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, badsnpname, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;
  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  for (i=0; i<numsnps; i++)  {  
	  cupt = snpmarkers[i] ;
	  if (cupt -> chrom == 23) cupt -> ignore = YES ;
  }

  if (poplistname != NULL) { 
   ZALLOC(eglist, numindivs, char *) ; 
  }
  else {
   setstatus(indivmarkers, numindivs, NULL) ;
   ZALLOC(eglist, MAXPOPS, char *) ;
   numeg = makeeglist(eglist, MAXPOPS, indivmarkers, numindivs) ;
  }
 for (i=0; i<numeg; i++) {  
  printf("%3d %s\n",i, eglist[i]) ;
 }

 nindiv=0 ;
 for (i=0; i<numindivs; i++) {
  indx = indivmarkers[i] ;
  if(indx -> affstatus == YES) ++nindiv  ;
 }

  if (noxdata) {
   for (i=0; i<numsnps; i++)  {  
    cupt = snpmarkers[i] ; 
    if (cupt -> chrom == 23) cupt-> ignore = YES ;
   }
  }

  rmsnps(snpmarkers, numsnps, nignore) ;  //  rid ignorable snps
  if (missingmode) setmiss(snpmarkers, numsnps) ;

  if  (weightname != NULL)   {  
   weightmode = YES ;
   getweights(weightname, snpmarkers, numsnps) ;
  }
   if (ldregress>0) {  
    ZALLOC(ldvv,  ldregress*numindivs, double) ;
    ZALLOC(ldvv2,  ldregress*numindivs, double) ;
    ZALLOC(vv2,  numindivs, double) ;
    ZALLOC(ldmat,  ldregress*ldregress, double) ;
    ZALLOC(ldmat2,  ldregress*ldregress, double) ;
    setidmat(ldmat, ldregress, 1.0e-6) ;
   }

  ZALLOC(xindex, numindivs, int) ;
  ZALLOC(xindlist, numindivs, Indiv *) ;
  ZALLOC(xsnplist, numsnps, SNP *) ;
  nrows = loadindx(xindlist, xindex, indivmarkers, numindivs) ;
  ncols = loadsnpx(xsnplist, snpmarkers, numsnps, indivmarkers) ;

  printf("## nrows: %d  ncols  %d\n", nrows, ncols) ;
  ZALLOC(xdata, nrows*ncols, double) ;
  ZALLOC(xmean, ncols, double) ;
  ZALLOC(xfancy, ncols, double) ;
// xfancy is multiplier for column xmean is mean to take off
  loadxdata(xdata, xsnplist, xindex, nrows, ncols) ;            
  if (rhomode) {  
   ZALLOC(tauvec, nrows, double) ;
   dotau(xdata, nrows, ncols, tauvec) ;
  }
  coladjust(xdata, nrows, ncols, fancynorm, xmean, xfancy, tauvec) ; 
/** MUST as xdata is raw until this call -1 => invalid */
  ZALLOC(XTX, nrows*nrows, double) ;
  ZALLOC(evecs, nrows*(nrows+rhonums+1), double) ;
  ZALLOC(tvecs, nrows*nrows, double) ;
  ZALLOC(cc, nrows + rhonums + 1 , double) ;
  ZALLOC(ww, nrows, double) ;

  lastchrom = -1 ;
  for (i=0; i<ncols; i++) { 
   cupt = xsnplist[i] ;
   chrom = cupt -> chrom ;
   getcol(cc, xdata, i, nrows, ncols) ;
   if (weightmode)  {
    weight = xsnplist[i] -> weight ;
    vst(cc, cc, xsnplist[i] -> weight, nrows) ;
   }
   if (ldregress>0) {  
      numclear = 0 ;
      for (k=1; k<= ldregress; ++k)  {  
       j = i-k ;  
       if (j<0) { 
        numclear = ldregress-k+1 ; 
        break ;
       }
       cupt2 = xsnplist[j] ;  
       if (cupt2 -> chrom != chrom) gdis = ldlimit + 1.0 ; 
       else gdis = cupt -> genpos - cupt2 -> genpos ;
       if (gdis>=ldlimit) {   
         numclear = ldregress-k+1 ; 
         break ;
       }
      }
      if (numclear>0) clearld(ldmat, ldvv, ldregress, nrows, numclear) ; 
      ldreg(ldmat, ldmat2, cc, vv2, ldvv, ldvv2, ldregress, nrows) ;
      copyarr(ldmat2, ldmat, ldregress*ldregress) ;
      copyarr(vv2, cc, nrows) ;
      copyarr(ldvv2, ldvv, ldregress*nrows) ;
    }

   addouter(XTX, cc, nrows) ;
   lastchrom = chrom ;
  }

  if (verbose) {
    
     if (tauvec == NULL) 
      printdiag(XTX, nrows) ;
     else 
      printdiagt(XTX, tauvec, nrows) ;

  }

  y = trace(XTX, nrows) / (double) (nrows-1) ;
  if (y<=0.0) fatalx("XTX has zero trace\n") ;
  vst(XTX, XTX, 1.0/y, nrows * nrows) ;

  eigvecs(XTX, cc, evecs, nrows) ;

   // cc are eigenvalues here
   m = numgtz(cc, nrows)  ;
  printf("matrix rank: %d\n", m) ;
  y = -1.0 ;
  printf("%8s %4s  %12s", "", "num", "eigenvalue") ;
  printf("%12s", "first diff") ;
  printf(" %9s %12s", "TW-stat" , "p-value") ;
  if (qtmode) printf(" %12s", "abs(qcoord)") ;
  printf(" %9s", "esize") ;
  printf("\n") ;
  if (qtmode) { 
   ZALLOC(qvec, nrows, double) ;
   ZALLOC(qcoord, nrows, double) ;
   for (j=0; j<nrows; j++) {  
    indx = xindlist[j] ;
    qvec[j] = indx -> qval ;
   }
   vadjust(qvec, nrows, NULL) ;
   y = asum2(qvec, nrows) ;  
   vst(qvec, qvec, 1.0/sqrt(y), nrows) ;
  }
  for (i=0; i<m; ++i) { 
    if ((m-i)>nostatslim) {
     tw = twestxx(cc+i, m-i, &zn, &zvar) ;
 //  printf("eignumber: %4d  effective marker size:  %12.3f\n", i, zn) ;
     tail = twtail(tw) ;
    }
    else { 
     zn = -1.0 ; 
     tail = -1.0 ; 
    }
    printf("%8s %4d  %12.6f","eigval:", i+1, cc[i]) ;
    if (i==0) printf( "%12s", "NA") ;
    else printf("%12.6f", cc[i]-cc[i-1]) ;
    if (tail>=0.0) printf(" %9.3f %12.6g", tw, tail) ;
    else printf(" %9s %12s", "NA", "NA") ;
    if (qtmode)  {  
     qcoord[i] = y = vdot(qvec,  evecs+i*nrows, nrows ) ;
     printf(" %12.6f", fabs(y)) ;
     chisq = y*y* (double) (nrows-1) ;
     tail = rtlchsq(1, chisq) ;
     printf (" %12.6g", tail) ;
    }
   if (zn>0.0) {
    printf(" %9.3f", zn) ;
   }
   else  {
    printf(" %9s", "NA") ;
   }
    printf("\n") ;
  }
  if (qtmode) return 0 ;
  numeigs = MIN(numeigs, nrows) ;
  if (numperms>0) {  
   seed = seednum() ;
   SRAND(seed) ;
   ZALLOC(numhits, numeigs, int) ;
   permtest(numhits, cc, numeigs, xdata, numperms, nrows, ncols, weightmode, xsnplist) ;
   for (i=0; i<numeigs; i++) { 
    t = numhits[i] ;
    printf("permtest: %3d %5d %5d\n", i, t, numperms-t) ;
   }
   return 0 ;
  }

  fprintf(ofile, "%20s ", "#eigvals:") ;
  for (j=0; j<numeigs; j++) { 
	  fprintf(ofile, "%9.3f ", cc[j]) ;
  }
  fprintf(ofile, "\n") ;

  ZALLOC(xtypes, nrows, int) ;

  ZALLOC(fvecs, nrows*numeigs, double) ;
  ZALLOC(fxvecs, nrows*numeigs, double) ;
  ZALLOC(fxscal, numeigs, double) ;

  ZALLOC(ffvecs, ncols*numeigs, double) ;
  ZALLOC(xrow, ncols, double) ;
  setfvecs(fvecs, evecs, nrows, numeigs) ;
  mulmat(ffvecs, fvecs, xdata, numeigs, nrows, ncols) ; // numeigs*ncols


  for (i=0; i<nrows; i++) {

   indx = xindlist[i] ;
   k = indxindex(eglist, numeg, indx -> egroup) ;
   xtypes[i] = k ;

   loadxdataind(xrow, xsnplist, xindex[i],  ncols) ;            
   fixxrow(xrow, xmean, xfancy, ncols) ;
   for (j=0; j<numeigs; j++) { 
     y = fxvecs[j*nrows+i] = vdot(xrow, ffvecs+j*ncols, ncols) ;
     fxscal[j] += y*y ;
   }
  }

  for (j=0; j<numeigs; j++) { 
    y = fxscal[j] ;                                            
    fxscal[j] = 10.0/sqrt(y) ; // eigenvectors have norm 10 (perhaps eccentric)
  }

  for (i=0; i < numindivs ; i++)  { 
          indx = indivmarkers[i] ;
          if (indx -> ignore) continue ;
          loadxdataind(xrow, xsnplist, i,  ncols) ;            
          fixxrow(xrow, xmean, xfancy, ncols) ;
	  fprintf(ofile, "%20s ", indx -> ID) ;
	  for (j=0; j<numeigs; j++) { 
           y = fxscal[j]*vdot(xrow, ffvecs+j*ncols, ncols) ;
           fprintf(ofile, "%9.3f  ", y) ;
	  }
	  fprintf(ofile, "%15s\n", indx -> egroup) ;
  }

  if (dotpopsmode == NO)  {
    printf("##end of run\n") ;
    printxcorr(XTX, nrows, xindlist) ;
    return 0 ;
  }

  ZALLOC(chitot, numeg*numeg, double) ;
  ZALLOC(xpopsize, numeg, int) ;

  dotpops(XTX, eglist, numeg, xtypes, nrows) ;


  for (j=0; j< numeigs; j++)  {  
   sprintf(sss, "eigenvector %d", j+1) ;
   y=dottest(sss, evecs+j*nrows, eglist, numeg, xtypes, nrows) ;
  }
  for (k1=0; k1<numeg; ++k1) {  
   for (k2=k1+1; k2<numeg; ++k2) {  
    ychi = chitot[k1*numeg+k2] ;
    tail = rtlchsq(numeigs, ychi) ;
    printf("popdifference:  %20s  %20s  %12.3f  %12.6g", eglist[k1], eglist[k2], ychi, tail) ;
    printf ("   %5d", xpopsize[k1]) ;
    printf ("   %5d", xpopsize[k2]) ;
    printf("\n") ;
   }
  }
  for (i=0; i<ncols; i++)  {  
   if (markerscore == NO) break;
   cupt = xsnplist[i] ;
   getcol(cc, xdata, i, nrows, ncols) ;
   sprintf(sss, "%s raw", cupt -> ID) ;
   dottest(sss, cc, eglist, numeg, xtypes, nrows) ;
   for (j=0; j< numeigs; j++)  {  
    sprintf(sss, "%s subtract sing vec %d", cupt ->ID, j+1) ;
    y = vdot(cc, evecs+j*nrows, nrows) ;
    vst(ww, evecs+j*nrows, y, nrows) ;
    vvm(cc, cc, ww, nrows) ;
    dottest(sss, cc, eglist, numeg, xtypes, nrows) ;
   }
  }

  printxcorr(XTX, nrows, xindlist)  ; 

  printf("##end of run\n") ;
  return 0 ;
}

void readcommands(int argc, char **argv) 

{
  int i,haploid=0;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   if (parname==NULL) { 
    fprintf(stderr, "no parameters\n") ;
    return ;
   }

   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "poplistname:", &poplistname) ;
   getstring(ph, "output:", &outputname) ;
   getstring(ph, "badsnpname:", &badsnpname) ;
   getstring(ph, "weightname:", &weightname) ;
   getint(ph, "numeigs:", &numeigs) ;
   getint(ph, "markerscore:", &markerscore) ; 
   getint(ph, "qtmode:", &qtmode) ; 
   getint(ph, "numperms:", &numperms) ; 
   getint(ph, "chisqmode:", &chisqmode) ; 
   getint(ph, "missingmode:", &missingmode) ; 
   getint(ph, "plotmode:", &plotmode) ; 
   getint(ph, "fancynorm:", &fancynorm) ; 
   getint(ph, "dotpopsmode:", &dotpopsmode) ; 
   getint(ph, "rhomode:", &rhomode) ;  /* correlation matrix */
   getint(ph, "pcorrmode:", &pcorrmode) ;  /* print correlations */
   getint(ph, "pcpopsonly:", &pcpopsonly) ;  /* but only within population */
   getint(ph, "noxdata:", &noxdata) ; 
   getint(ph, "nostatslim:", &nostatslim) ; 

   getint(ph, "ldregress:", &ldregress) ;
   getdbl(ph, "ldlimit:", &ldlimit) ;  /* in morgans */


   printf("### THE INPUT PARAMETERS\n");
   printf("##PARAMETER NAME: VALUE\n");
   writepars(ph);

}
int
loadindx(Indiv **xindlist, int *xindex, Indiv **indivmarkers, int numindivs) 
{
  int i, n=0 ;  
  Indiv *indx ;
  for (i=0; i<numindivs; i++)  {
   indx = indivmarkers[i] ;
   if (indx -> ignore) continue ;
   if (indx -> affstatus == NO) continue ;
   xindex[n] = i ;
   xindlist[n] = indx ;
   ++n ;
  }
  return n ;
}
int
loadsnpx(SNP **xsnplist, SNP **snpmarkers, int numsnps, Indiv **indivmarkers) 
{
  int i, n=0 ;  
  SNP *cupt ;
  for (i=0; i<numsnps; i++)  {
   cupt = snpmarkers[i] ;
   if (cupt -> ignore) continue ;
   if (numvalidgt(indivmarkers, cupt) == 0) continue ;
   xsnplist[n] = cupt ;
   ++n ;
  }
  return n ;
}
void
loadxdata(double *xdata, SNP **snplist, int *xindex, int nrows, int ncols)             
{
 SNP *cupt ;
 Indiv *indx ;
 int i, j, k, n,  g ;

 for (i=0; i<ncols; i++) {  
  cupt = snplist[i] ;
  n = cupt -> ngtypes ;
  if (n<nrows) fatalx("bad snp: %s %d\n", cupt -> ID, n) ;
  for (j=0; j< nrows; j++) {  
   g = cupt ->  gtypes[xindex[j]] ;
   k = j*ncols + i ; 
   xdata[k] = (double) g ;
  }
 }
}
void
loadxdataind(double *xrow, SNP **snplist, int ind,  int ncols)             
{
 SNP *cupt ;
 Indiv *indx ;
 int i, j, k, n,  g ;

 for (i=0; i<ncols; i++) {  
   cupt = snplist[i] ;
   g = cupt ->  gtypes[ind] ;
   xrow[i] = (double) g ;
 }
}
void fixxrow(double *xrow, double *xmean, double *xfancy, int len) 
{
    int i ;

    vvt(xrow, xrow, xfancy, len) ;
    for (i=0; i<len; i++) { 
     if (xrow[i] < -0.1) xrow[i] = 0.0 ; 
     else xrow[i] -= xmean[i] ;
    }
}

void coladjust(double *xdata, int nrows, int ncols, 
 int fancy, double *xm, double *xf, double *tauv)  
// deal with missing data force missing 0 
{
  int i, nmiss ;
  double *cc, y, y2, mean ;
  double *xx ;  

  ZALLOC(cc, nrows, double) ;
  ZALLOC(xx, nrows, double) ;
  if (tauv == NULL) {  
   vclear(xx, 1.0, nrows) ;  
  }
  else { 
    for (i=0; i<nrows; i++) { 
     y = tauv[i] ;
     xx[i] = 1.0/sqrt(1.0-y) ;    
    }
  }

  vclear(xf, 1.0, ncols) ;
  for (i=0; i<ncols; i++) { 
   getcol(cc, xdata, i, nrows, ncols) ;
   vvt(cc, cc, xx, nrows) ;
   if (fancy) dofancy(cc, nrows, xf+i) ;
   nmiss = vadjust(cc, nrows, xm+i) ;
   putcol(cc, xdata, i, nrows, ncols) ;
  }
  free(cc) ;
  free(xx) ;
}
void dofancy(double *cc, int n, double *fancy) 
{
  int i, t, nmiss=0 ;
  int top, bot ;  
  double p, yvar, y ;

  top = bot = 0 ;
  for (i=0; i<n; i++) {  
   t = nnint(cc[i]) ;
   if (t<0) { 
    ++nmiss ;
    continue ;
   }
   top += t ;
   bot += 2 ;
  }
  if (bot==0) return ;
  if (top == 0) return ;
  if (top == bot) return ;
  p = (double) top / ((double) bot) ;
  yvar = p*(1.0-p) ;
  y = 1.0/sqrt(yvar) ;
  vst(cc, cc, y, n )  ;                    
  *fancy = y ;
}

int vadjust(double *cc, int n, double *pmean) 
/* take off mean  force missing to zero */
{
 double ynum, ysum, y, ymean ;
 int i, nmiss=0 ;

 ynum = ysum = 0.0 ;
 for (i=0; i<n; i++) {  
  y = cc[i] ;
  if (y < 0.0) { 
   ++nmiss ;
   continue ;
  }
  ++ynum ;
  ysum += y ;
 }
 if (ynum==0.0) fatalx("logic bug all missing\n") ;
 ymean = ysum/ynum ;
 for (i=0; i<n; i++) {  
  y = cc[i] ;
  if (y < 0.0) cc[i] = 0.0 ;
  else cc[i] -= ymean ;
 }
 if (pmean != NULL) *pmean = ymean ;
 return nmiss ;
}

void
getcol(double *cc, double *xdata, int col, int nrows, int ncols) 
{
 int i ;

 for (i=0; i<nrows; i++) { 
  cc[i] = xdata[i*ncols+col] ;
 }
}

void
putcol(double *cc, double *xdata, int col, int nrows, int ncols) 
{
 int i ;

 for (i=0; i<nrows; i++) { 
  xdata[i*ncols+col] = cc[i] ;
 }
}
double
dottest(char *sss, double *vec, char **eglist, int numeg, int *xtypes, int len) 
// vec will always have mean 0 
// perhaps should rewrite to put xa1 etc in arrays
{
   double *w1 ; 
   int *xt ;
   int i, k1, k2, k, j, n ;
   double y1, y2, ylike, yl0, yl1, yl2 ;
   double ychi ;
   double *wmean ;
   int imax, imin ;
   static int ncall = 0 ;

   char ss1[MAXSTR] ;
   char ss2[MAXSTR] ;
   char sshit[4] ;
   double tail, ans, ftail, ftailx, ansx ; 

   ZALLOC(wmean, numeg, double) ;
   ZALLOC(w1, len, double) ;
   ZALLOC(xt, len, int) ;
   strcpy(ss1, "") ;

   calcmean(wmean, vec, len, xtypes, numeg) ;
   vlmaxmin(wmean, numeg, &imax, &imin) ;  
    if (chisqmode) {
     ylike = anova1(vec, len, xtypes, numeg) ;
     ans = 2.0*ylike ;
    }
    else {
     ans = ftail = anova(vec, len, xtypes, numeg) ;
    }
    ++ncall ;

    
    if (numeg>2) {
     sprintf(ss2, "%s %s ", sss, "overall") ;
     publishit(ss2, numeg-1, ans) ;
     printf(" %20s minv: %9.3f %20s maxv: %9.3f\n", 
     eglist[imin], wmean[imin], eglist[imax], wmean[imax]) ;
    }


    for (k1 = 0; k1<numeg; ++k1) { 
     for (k2 = k1+1; k2<numeg; ++k2) { 
      n = 0 ;
      for (i=0; i<len ; i++)   {  
        k = xtypes[i] ;
        if (k == k1) {  
         w1[n] = vec[i] ; 
         xt[n] = 0 ; 
         ++n ;
        }
        if (k == k2) {  
         w1[n] = vec[i] ; 
         xt[n] = 1 ; 
         ++n ;
        }
      }

     ylike = anova1(w1, n, xt, 2) ;
     ychi  = 2.0*ylike ;
     chitot[k1*numeg + k2]  += ychi ;
     if (chisqmode) {
      ansx = ychi ;
     }
     else {
      ansx = ftailx = anova(w1, n, xt, 2) ;
     }

      sprintf(ss2,"%s %s %s ", sss, eglist[k1], eglist[k2]) ;
      publishit(ss2, 1, ansx) ;
      
     }
    }
    free(w1) ;
    free(xt) ;
    free(wmean) ;
    return ans ;
}
double yll(double x1, double x2, double xlen) 
{
  double m1, m2, var ;

  if (xlen < 1.5 ) return 0.0 ;
  m1 = x1/xlen ;
  m2 = x2/xlen ;  
  var = m2-m1*m1 ;
  if (var <= 0.0) fatalx("bad yll\n") ;
  return -0.5*xlen*log(var) ;
}
void
calcmean(double *wmean, double *vec, int len, int *xtypes, int numeg) 
{

   int i, k ;
   double y1 ;
   double *w0, *popsize ;

   ZALLOC(w0, len, double) ;
   ZALLOC(popsize, numeg, double) ;

   y1 = asum(vec, len)/ (double) len ;  // mean
   vsp(w0, vec, -y1, len) ;

    for (i=0; i<len; i++)  { 
     k = xtypes[i] ;
     ++popsize[k] ;
     wmean[k] += w0[i] ;
    }



    vsp(popsize, popsize, 1.0e-12, numeg) ;
    vvd(wmean, wmean, popsize, numeg) ;

    free(w0) ;
    free(popsize) ;



}

double anova(double *vec, int len, int *xtypes, int numeg)
// anova 1 but f statistic
{
   int i, k ; 
   double y1, y2, ylike, top, bot, ftail ;  
   double *w0, *w1, *popsize, *wmean ;

   static int ncall2  = 0 ;

   if (numeg >= len) fatalx("bad anova\n") ;
   ZALLOC(w0, len, double) ;
   ZALLOC(w1, len, double) ;
   ZALLOC(wmean, numeg, double) ;
   ZALLOC(popsize, numeg, double) ;

   y1 = asum(vec, len)/ (double) len ;  // mean
   vsp(w0, vec, -y1, len) ;

    for (i=0; i<len; i++)  { 
     k = xtypes[i] ;
     ++popsize[k] ;
     wmean[k] += w0[i] ;
    }

/* debug */
    if (numeg == 2)  {  
     ++ncall2 ;
     for (i=0; i<len; ++i) {  
      if (ncall2<0) break ;
      k = xtypes[i] ;
//    printf("yy %4d %4d %12.6f %12.6f\n", i, k, vec[i], w0[i]) ;
     }
    }

    vsp(popsize, popsize, 1.0e-12, numeg) ;
    vvd(wmean, wmean, popsize, numeg) ;

    vvt(w1, wmean, wmean, numeg) ;
    top = vdot(w1, popsize, numeg) ;
    
    for (i=0; i<len ; i++)   {  
     k = xtypes[i] ;
     w1[i] = w0[i] - wmean[k] ;
    }
    bot = asum2(w1, len) / (double) (len-numeg) ;
    bot *= (double) (numeg-1) ;
//  printf("zz anova: %4d  %15.9f  %15.9f\n", numeg-1, top, bot) ;
    ftail = rtlf(numeg-1, len-numeg, top/bot) ;

    free(w0) ; 
    free(w1) ; 
    free(popsize) ;
    free(wmean) ;

    return ftail ;

}
double anova1(double *vec, int len, int *xtypes, int numeg)
{
   int i, k ; 
   double y1, y2, ylike ;  
   double *w0, *w1, *popsize, *wmean ;

   ZALLOC(w0, len, double) ;
   ZALLOC(w1, len, double) ;
   ZALLOC(wmean, numeg, double) ;
   ZALLOC(popsize, numeg, double) ;

   y1 = asum(vec, len)/ (double) len ;  // mean
   vsp(w0, vec, -y1, len) ;

    for (i=0; i<len; i++)  { 
     k = xtypes[i] ;
     ++popsize[k] ;
     wmean[k] += w0[i] ;
    }

    vsp(popsize, popsize, 1.0e-12, numeg) ;
    vvd(wmean, wmean, popsize, numeg) ;
    
    for (i=0; i<len ; i++)   {  
     k = xtypes[i] ;
     w1[i] = w0[i] - wmean[k] ;
    }

    y1 = asum2(w0, len) / (double) len ;
    y2 = asum2(w1, len) / (double) len ;
    ylike = 0.5*((double) len)*log(y1/y2) ;

    free(w0) ; 
    free(w1) ; 
    free(popsize) ;
    free(wmean) ;

    return ylike ;

}
void publishit(char *sss, int df, double chi) 
{
      double tail ;
      char sshit[4] ;
      char ss2[MAXSTR] ;
      int i, n ;
      char cblank, cunder ;
      static int ncall = 0 ;

      ++ncall ;
      cblank = ' ' ;
      cunder = '_' ;
      n = strlen(sss) ;

      strcpy(ss2, sss) ;  
      for (i=0; i< n; ++i)  { 
       if (ss2[i] == cblank)  ss2[i] = cunder ;
      }

      if (chisqmode) {
       if (ncall==1) printf("%40s %6s %9s %12s\n", "", "dof", "chisq", "p-value") ;
       printf("%40s %6d %9.3f",ss2, ss2, df, chi) ;
       tail = rtlchsq(df, chi) ;  
       printf(" %12.6g", tail) ;
      }
      else {  
       if (ncall==1) printf("%40s %12s\n", "",  "p-value") ;
       printf("%40s ", ss2) ;
       tail = chi ;  
       printf(" %12.6g", tail) ;
      }
      strcpy(sshit, "") ;  
      if (tail < pvhit) strcpy(sshit, "***") ;
      if (tail < pvjack) strcpy(sshit, "+++") ;
      printf(" %s", sshit) ;
      printf("\n") ;
}

void
permtest(int *numhits, double *evals, int numeigs, double *xdata, 
 int numperms, int nrows, int ncols, int weightmode, SNP **xsnplist) 
{
       double *XTX, *pevals, *cc ;
       int iter, k, i ;
       double weight, y ;

      ZALLOC(pevals, nrows, double) ;
      ZALLOC(cc, nrows, double) ;
      ZALLOC(XTX, nrows*nrows, double) ;

      for (iter=1; iter<=numperms; ++iter) {  
       vzero(XTX, nrows*nrows) ;
       for (i=0; i<ncols; i++) { 
        getcol(cc, xdata, i, nrows, ncols) ;
        if (weightmode)  {
        weight = xsnplist[i] -> weight ;
        vst(cc, cc, xsnplist[i] -> weight, nrows) ;
        }
        randperm(cc, nrows) ;
        addouter(XTX, cc, nrows) ;
       }
       y = trace(XTX, nrows) / (double) (nrows-1) ;
       if (y<=0.0) fatalx("XTX has zero trace\n") ;
       vst(XTX, XTX, 1.0/y, nrows * nrows) ;
       eigvals(XTX, pevals, nrows) ;
       for (k=0; k<numeigs; ++k) {  
        if (pevals[k]>evals[k]) ++numhits[k] ;
       }
      }

      free(XTX) ;
      free(pevals) ;
      free(cc) ;
}
void randperm(double *x, int n) 
{
    int *ind ;

    ZALLOC(ind, n, int) ;

    idperm(ind, n) ;
    ranperm(ind, n) ; 
    dpermute(x, ind, n) ;

    free(ind) ;
}
void
setmiss(SNP **snpm, int numsnps) 
{
   SNP *cupt ;
   int i, j, t, n, tot ;

   for (i=0; i<numsnps; i++)  {  
    cupt = snpm[i] ;
    n = cupt -> ngtypes ;
    if (n <= 0) continue ;
    tot = 0 ;
    for (j=0; j<n; j++) {  
     if (cupt -> gtypes[j] >= 0) { 
      t = 1 ;
     }
     else { 
      t = 0 ;
     }
     cupt -> gtypes[j] = t ;
     tot +=t ;
    }
    if (verbose)
     printf("Valids: %s %d\n", cupt -> ID, tot) ;
   }
   

}
void
setfvecs(double *fvecs, double *evecs, int nrows, int numeigs) 
// plotmode each eigenvector min 0 max 1
{

   double *w ;
   double xmax, xmin ;
   int i, j ;

   ZALLOC(w, nrows, double) ;

   for (j=0; j<numeigs; j++) {  
    copyarr(evecs+j*nrows, w, nrows) ;
    if (plotmode==NO) {   
     vst(fvecs+j*nrows, w, 10.0, nrows) ;
     continue ;
    }
    if (maxmode) {
     vmaxmin(w, nrows, &xmax, &xmin) ;
     vsp(w, w, -xmin, nrows) ; //  min now zero 
     xmax -= xmin ;
     vst(w, w, 1.0/xmax, nrows) ;
    }
    copyarr(w, fvecs+j*nrows, nrows) ;
   }
   free(w) ;
}
void
dotpops(double *X, char **eglist, int numeg, int *xtypes, int nrows) 
{
      double *pp, *npp, val, yy ;
      int *popsize ;
      int i, j, k1, k2 ;


     ZALLOC(pp, numeg * numeg, double) ;
     ZALLOC(npp, numeg * numeg, double) ;
     popsize = xpopsize; 

     for (i=0; i<nrows; i++) { 
      k1 = xtypes[i] ;
      ++popsize[k1] ;
      for (j=i+1; j<nrows; j++) { 
       k2 = xtypes[j] ;
       val = X[i*nrows+i] + X[j*nrows+j] - 2.0*X[i*nrows+j] ;
       pp[k1*numeg+k2] += val ;
       pp[k2*numeg+k1] += val ;
       ++npp[k1*numeg+k2]  ;
       ++npp[k2*numeg+k1]  ;
      }
     }
     vsp(npp, npp, 1.0e-8, numeg*numeg) ;
     vvd(pp, pp, npp, numeg*numeg) ;
// and normalize so that mean on diagonal is 1 
     yy = trace(pp, numeg) / (double) numeg ;
     vst(pp, pp, 1.0/yy, numeg*numeg) ;
     if (numeg<=10) {
      printf("\n") ;
      printf("%10s", "") ;
      for (k1=0; k1<numeg; ++k1) {  
       printf(" %10s", eglist[k1]) ;
      }
      printf("  %10s", "popsize") ;
      printf("\n") ;
      for (k2=0; k2<numeg; ++k2) {  
       printf("%10s", eglist[k2]) ;
       for (k1=0; k1<numeg; ++k1) {  
        val = pp[k1*numeg+k2] ;
        printf(" %10.3f", val) ;
       }
       printf("  %10d", popsize[k2]) ;
       printf("\n") ;
      }
     }
     else {   // numeg >= 10 
      for (k2=0; k2<numeg; ++k2) {  
       for (k1=k2; k1<numeg; ++k1) {  
        printf("dotp: %10s", eglist[k2]) ;
        printf(" %10s", eglist[k1]) ;
        val = pp[k1*numeg+k2] ;
        printf(" %10.3f", val) ;
        printf("    %10d", popsize[k2]) ;
        printf(" %10d", popsize[k1]) ;
        printf("\n") ;
       }
      }
     }
    printf("\n") ;
    printf("\n") ;
     

    free(pp) ;
    free(npp) ;

}
void printxcorr(double *X, int nrows, Indiv **indxx) 
{
   int k1, k2, t ; 
   double y1, y2, yy, rho ;
   Indiv *ind1, *ind2 ;

   if (pcorrmode == NO) return ;
   for (k1=0; k1<nrows; ++k1) {  
    for (k2=k1+1; k2<nrows; ++k2) {  

     ind1 = indxx[k1] ;
     ind2 = indxx[k2] ;

     t = strcmp(ind1 -> egroup, ind2 -> egroup) ;
     if (pcpopsonly && (t != 0))  continue ;
     

     y1 = X[k1*nrows+k1] ;
     y2 = X[k2*nrows+k2] ;
     yy = X[k1*nrows+k2] ;

     rho = yy/sqrt(y1*y2+1.0e-20) ;
     printf("corr: %20s %20s %20s %20s %9.3f\n", 
      ind1 -> ID, ind2 -> ID, ind1 -> egroup, ind2 -> egroup, rho) ;

    }
   }
}
void clearld(double *ldmat, double *ldvv, int rsize, int n, int nclear)    
{
  int i, j, lo, hi ;

  if (nclear>rsize) fatalx("bad nclear\n") ;

  lo = rsize-nclear ;  
  hi = rsize-1 ;

  for (i=lo; i<=hi; ++i)  { 
   vzero(ldvv+i*n, n) ;
    for (j=0; j<=hi; ++j)  { 
      ldmat[j*rsize+i] = ldmat[i*rsize+j] = 0.0 ;
    }
// force matrix non-singular 
    ldmat[i*rsize+i] = 1.0e-8 ;  
  }
}

void
ldreg(double *ldmat, double *ldmat2, double *vv, double *vv2, double *ldvv, 
  double *ldvv2, int rsize, int n) 
/** ldmat2 is inner product matrix for last rsize columns on exit */
{
   int i, j, k1, k2 ; 
   double *rr, *ans, *tt ;
   double y ;

   ZALLOC(rr, rsize, double) ;
   ZALLOC(ans, rsize, double) ;
   ZALLOC(tt, n, double) ;

   if (rsize>1) 
    copyarr(ldvv, ldvv2+n, n*(rsize-1)) ;
   for (i=0; i<rsize-1 ; i++)  { 
    for (j=0; j<rsize-1 ; j++)  { 
     k1 = i*rsize+j  ;
     k2 = (i+1)*rsize+j+1 ;
     ldmat2[k2] = ldmat[k1] ;
    }
   }
   copyarr(vv, ldvv2, n) ;
   i = 0 ;
   for (j=0; j<rsize ; j++)  { 
     y = rr[j] = vdot(vv, ldvv+j*n, n) ;
     y =  vdot(vv, ldvv2+j*n, n) ;
     if (j==0) y += 1.0e-6 ;
     ldmat2[i*rsize+j] = ldmat2[j*rsize+i] = y ;
   }
  solvit(ldmat, rr, rsize, ans) ; /* solve normal equations */
  copyarr(vv, vv2, n) ;
  for (i=0; i<rsize; i++) {  
   vst(tt, ldvv+i*n, -ans[i], n) ;
   vvp(vv2, vv2, tt, n) ;
  }
  free(rr) ;
  free(ans) ;
  free(tt) ;
/**
  printf("zz %9.3f %9.3f\n", asum(vv, n), asum(vv2,n) ) ;
  printf("zz2 %9.3f %9.3f\n", asum2(vv, n), asum2(vv2,n) ) ;
*/
}
void
setidmat(double *mat, int n, double scal) 
{
  int i ;
  
  vzero(mat, n*n) ; 
  for (i=0; i<n; i++) { 
   mat[i*n+i] = scal ;
  }
}
void fixrho (double *a, int n) 
// turn a into correlation matrix
{
   double *d, *tt, y ;
   int i ;  

   ZALLOC(d, n, double) ;
   ZALLOC(tt, n*n, double) ;

   getdiag(d, a, n) ;

   vsqrt(d, d, n) ;
   addouter(tt, d, n) ;

   vvd(a, a, tt, n*n) ;


  free(d) ;
  free(tt) ;


}
void printdiag(double *a, int n) 
{
   double *d, *tt, y ;
   int i ;  

   ZALLOC(d, n, double) ;
   getdiag(d, a, n) ;
    y = asum(d,n) / (double) n ; 
    for (i=0; i<n; i++) {  
     printf("diag: %9.3f\n", d[i]/y) ;
    }


   free(d) ;

}
void printdiagt(double *a, double *xtau, int n) 
{
   double *d, *tt, y, y2 ;
   int i ;  

   ZALLOC(d, n, double) ;
   getdiag(d, a, n) ;
    y  = asum(d,n) / (double) n ; 
    y2 = asum(xtau,n) / (double) n ; 
    for (i=0; i<n; i++) {  
     printf("diagt: %9.3f  %9.3f %9.3f\n", d[i]/y, xtau[i]/y2, xtau[i]) ;
    }


   free(d) ;

}
void
dotau(double *xdata, int nrows, int ncols, double *tauvec) 

{
  int i, j, t, nmiss=0 ;
  int top, bot ;  
  double p,  y ;
  double *cc, *pp, *tt ;

  ZALLOC(cc, nrows, double) ;
  ZALLOC(pp, ncols, double) ;
  ZALLOC(tt, ncols, double) ;

  for (i=0; i<ncols; i++) {  
   getcol(cc, xdata, i, nrows, ncols) ;
   top = bot = 0 ;
   for (j=0; j<nrows; ++j)  {
    t = nnint(cc[j]) ;
    if (t<0) { 
     ++nmiss ;
     continue ;
    }
    top += t ;
    bot += 2 ;
   }
   if (bot == 0) bot = 1 ;
   p = (double) top / ((double) bot) ;
   pp[i]  = p ;  
 }
for (j=0; j<nrows; ++j)  {
 for (i=0; i<ncols; i++) {  
   tt[i] = xdata[j*ncols+i] ;
  }
/* call esttau etc */
  y = esttau(pp, tt, ncols) ;
  tauvec[j] = MAX(y, 0.0) ;
  printf("zzz3 %5d %9.3f\n", i, y) ;
 }

  free(cc) ; 
  free(pp) ;
  free(tt) ; 
}
double 
esttau(double *pp, double *cc, int n) 
{
   double *theta, *zz, znum, zden, p, q, tau ;  
   int i, k ;
   int iter, numiter = 10 ;
   double ylike, ybase ;
   double gg[2], y0, y1 ;  

   tau = easyesttau(pp, cc, n)  ; 
   tau = MAX(tau, 0.01) ;
   for (iter = 1; iter <= numiter; ++iter)  {
    ylike = 0.0 ;
    vzero(gg, 2) ;
    for (i=0; i<n; i++)  {  
     k = nnint(cc[i]) ;  
     p = pp[i] ; 
     q = 1.0-p ;

     switch (k) {  
      case 0:  
       y0 = q*q ;
       y1 = p*q*tau ;  
       gg[1] += y1/(y0+y1) ;
       ylike += log(y0+y1) ;
       break ;

      case 1:  
       y1 = 2*p*q*(1.0-tau) ;
       gg[0] += 1.0;
       ylike += log(y1) ;
       break ;

      case 2:  
       y0 = p*p ;
       y1 = p*q*tau ;  
       gg[1] += y1/(y0+y1) ;
       ylike += log(y0+y1) ;
       break ;
     }
    }
     if (iter==1) ybase = ylike ; 
     ylike -= ybase ;
//   printf("iter: %3d %9.3f %9.3f\n", iter, tau, ylike) ;
     tau = gg[1]/asum(gg,2) ;
   }
   return tau ;
}
double 
easyesttau(double *pp, double *cc, int n) 
{
   double *theta, *zz, znum, zden, p, tau ;  
   int i, k ;

   ZALLOC(zz, n, double) ;
   ZALLOC(theta, n, double) ;

   for (i=0; i<n; i++) {  
    p = pp[i] ;

    theta[i] = 2*p*(1-p) ;
    k = nnint(cc[i]) ;
    if (k==1) zz[i] = 1 ;

  }
  znum = asum(zz, n) ;
  zden = asum(theta, n) ;  
  printf("zsum: %12.3f  %9.3f\n", znum, asum(zz, n)) ;
  tau = 1.0 - (znum/zden) ;   
  printf ("tau: %12.6f\n", tau) ;
  free(zz) ; 
  free(theta) ;
  return tau ;
} 

