#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <nicklib.h>
#include <getpars.h>

#include "admutils.h"
#include "mcio.h"  

#define WVERSION   "1001" 
// extra space in output for long names

/** write haplotypes in nick format */

#define MAXFL  50   
#define MAXSTR  512

char *trashdir = "/var/tmp" ;

int verbose = NO ;
Indiv **indivmarkers;
SNP **snpmarkers ;
int numsnps, numindivs ; 

char  *genotypename = NULL ;
char  *snpname = NULL ;
char  *indivname = NULL ;
char *outputname = NULL ;
char *auxname = NULL ;
FILE *ofile, *auxfile ;
int  lopos, hipos, xchrom ;

void readcommands(int argc, char **argv) ;

void gethpos(int *fc, int *lc, SNP **snpm, int numsnps, int xchrom, int lo, int hi) ;
void sethap(char *hap, int *lenhap, SNP **snpm,  Indiv *indx,  
  int fc, int lc)  ;


int main(int argc, char **argv)
{

  int **snppos ;
  int *snpindx ;
  char **snpnamelist, **indnamelist ;
  char **eglist ;
  char **aglist ;
  int  lsnplist, lindlist, numeg, numag ;
  int i,j,k; 
  SNP *cupt, *cupt1, *cupt2 ;
  Indiv *indx ;
  char str[MAXSTR] ;

  int nindiv = 0, e, f, lag=1  ;
  int n, n0, n1, g, chrom, gender ;
  int nignore, numrisks = 1;
  int fc, lc, lenhap  ;
  char *hap ; 

 readcommands(argc, argv) ;

  numsnps = 
    getsnps(snpname, &snpmarkers, 0.0, NULL, &nignore, numrisks) ;

  numindivs = getindivs(indivname, &indivmarkers) ;
  getgenos(genotypename, snpmarkers, indivmarkers, 
    numsnps, numindivs, nignore) ;

  openit (outputname, &ofile, "w") ;
  gethpos(&fc, &lc, snpmarkers, numsnps, xchrom, lopos, hipos)   ;


  if (fc > numsnps) fatalx("bad fc %d\n", fc) ;
  ZALLOC(hap, lc-fc+10, char) ; 
  for (i=0; i<numindivs ; i++)  {  
    indx = indivmarkers[i] ;
    if (indx -> ignore) continue ;
    sethap(hap, &lenhap, snpmarkers,  indx,  fc, lc) ; 
    fprintf(ofile, "%s %4d %d ", hap, 2*i, 0) ;
    fprintf(ofile, "%20s %20s %c\n", indx->ID, indx -> egroup, 'A')  ;
    fprintf(ofile, "%s %4d %d ", hap, 2*i+1, 0) ;
    fprintf(ofile, "%20s %20s %c\n", indx->ID, indx -> egroup, 'B')  ;
  }
  fclose(ofile)  ;
  if (auxname != NULL) { 
   openit(auxname, &auxfile, "w") ;
   for (k=fc ; k <=lc; ++k) {  
    cupt = snpmarkers[k] ;
    fprintf(auxfile, "%3d %20s %12d\n", k-fc, cupt -> ID, nnint(cupt -> physpos)) ;
   }
   fclose (auxfile) ;
  }
  printf("===indivs===\n") ;
  for (i=0; i<numindivs; i++) {  
   indx = indivmarkers[i] ;
   n = numvalids(indx, snpmarkers, fc, lc) ;
   printf("%20s %5d\n", indx -> ID, n) ;
  }
  printf("===snps===\n") ;
  for (i=fc; i <= lc; i++) {  
   cupt = snpmarkers[i] ;
   n = numvalidgtypes(cupt) ;
   printf("%20s %5d\n", cupt -> ID, n) ;
  }

  return 0 ;

}

void readcommands(int argc, char **argv) 
{
  int i,haploid=0;
  char *parname = NULL ;
  phandle *ph ;
  char str[5000]  ;
  char *tempname ;
  int n ;

  while ((i = getopt (argc, argv, "p:vV")) != -1) {

    switch (i)
      {

      case 'p':
	parname = strdup(optarg) ;
	break;

      case 'v':
	printf("version: %s\n", WVERSION) ; 
	break; 

      case 'V':
	verbose = YES ;
	break; 

      case '?':
	printf ("Usage: bad params.... \n") ;
	fatalx("bad params\n") ;
      }
  }

         
   pcheck(parname,'p') ;
   printf("parameter file: %s\n", parname) ;
   ph = openpars(parname) ;
   dostrsub(ph) ;

   getstring(ph, "genotypename:", &genotypename) ;
   getstring(ph, "snpname:", &snpname) ;
   getstring(ph, "indivname:", &indivname) ;
   getstring(ph, "output:", &outputname) ;
   getstring(ph, "auxname:", &auxname) ;
   getint(ph, "lopos:", &lopos) ;
   getint(ph, "hipos:", &hipos) ;
   getint(ph, "chrom:", &xchrom) ;

}

void sethap(char *hap, int *lenhap, SNP **snpm,  Indiv *indx,  
  int fc, int lc) 
/** return haplotypes string */
{

  char *ss ;   
  int  *xhap ;
  SNP *cupt ;
  int k, n, val, oval, len ;
  int indiv ;

  len = lc-fc+1 ;
  ZALLOC(ss, len+1, char) ;
  ZALLOC(xhap, len, int) ;

  indiv = indx -> idnum ;  
  for (k=fc; k<= lc; k++)  {  
   cupt = snpm[k] ;
   n = cupt -> ngtypes ;
   if (n<=indiv) {
     if (indiv==0) 
      printf("***warning: no snp data: %s %d\n", cupt -> ID, n) ;
     val = -1 ;
   }
   else {
    val = cupt -> gtypes[indiv] ;
   }
   oval = 0 ;
   if (val==0) oval = 1 ;
   if (val==1) oval = 9 ;  // het
   if (val==2) oval = 2 ;  // het
   xhap [k-fc] = oval ;
  }
 for (k=0; k< len; k++) {  
  sprintf(ss+k, "%d", xhap[k]) ;
 }
 strcpy(hap, ss) ;
 *lenhap = len ;       

 free(ss) ;
 free(xhap) ;

}
