#include <stdio.h>
#include <limits.h>
#include <math.h>
#include "nicklib.h" 

#define N  10000
#define MAXSTR 512
#define MAXFF 50

int readit(double *xx) ;
void probit(double *xout, double *xin, int n)  ;

/** 
 takes a single column file and     
 1) Computes a Z-statistic  
 2) does probit transform  
 Outputs (data, Z, probit)
*/

int main(void)
{
   double xx[N] ; 
   double *a, *b ;
   double y ;
   int n, i ;
  
   n = readit(xx) ;
   ZALLOC(a,n, double) ;
   ZALLOC(b,n, double) ;
   y = asum(xx, n) / (double) n ;
   vsp(a, xx, -y, n) ;
   y = asum2(a, n) / (double) n ;
   y = sqrt(y) ;
   vst(a, a, 1.0/y, n) ;  // Z statistic
   
// now dither at random
   gaussa(b, n) ;
   vst(b, b, 1.0e-6, n) ;
   vvp(b, a, b, n) ;
   probit(b, b, n) ;

   for (i=0; i<n; i++) { 
    printf("%9.3f %9.3f %9.3f\n", xx[i], a[i], b[i]) ;
   }
}

int 
readit(double *xx)
{

  char line[MAXSTR] ;
  char *spt[MAXFF] ;
  char *sx ;
  int nsplit, i, j, num=0 ;

  while (fgets(line, MAXSTR, stdin) != NULL)  {
   nsplit = splitup(line, spt, MAXFF) ; 
   sx = spt[0] ;
   if (sx[0] == '#') {
    freeup(spt, nsplit) ;
    continue ;
   }
   if (nsplit<1) { 
     continue ;
   }
   if (num>=N) fatalx("too much data\n") ;
   xx[num] = atof(spt[0]) ;
   freeup(spt, nsplit) ;
   ++num ;
  }
  return num ;
}

void probit(double *xout, double *xin, int n) 
/** probit transform */
{
   int i ;
   double tail, y, top, bot ; 
   double *a ;
   int *ind ;

   ZALLOC(a, n, double) ;
   ZALLOC(ind, n, int) ;

   vst(a, xin, -1.0, n) ;
   sortit(a, ind, n) ;
   bot = (double) (n+1) ;
   for (i=0; i<n; i++)  {
    top = (double) (i+1) ;
    tail = top/bot ;
    y = zprob(tail) ;
    xout[ind[i]] = y ;
   }

/**
   for (i=0; i<n; i++)  {
    printf("zz %4d %9.3f %9.3f\n", i, xin[i], xout[i]) ;
   }
*/

   free(a) ;
   free(ind) ;


}

