#include <stdio.h>
#include <limits.h>
#include <math.h>  
#include <nicklib.h> 
#include "admutils.h"  

int verbose = NO ;

/* ********************************************************************* */
char *trashdir = NULL ; 
int qtmode = NO ;

#define  NTEST
/* 
 a testdriver for mutxhmm which is used in the latest ranksnp
*/

#define  N  8  
double
 mutxhmmd(double *p0, double *p1, double *dis, int n, int kmark, double theta, double lambda)  ;
double 
mutxhmm(double *p0, double *p1, double *dis, int n, int kmark, double theta, double lambda)  ;
void flipx (double *a, double *b, int n)  ;
void calcM(double poisson, double theta, double **M, int numstates) ;
void calcx(double *X, double PopA_freq, double PopB_freq, int gtype, int numstates)  ;
void decode(int *gtype, int pat, int l, int base) ;

int main(void)
{
  double p0[N], p1[N], dis[N] ;   
  double theta = 0.2, lambda = 6.0 ;
  double yd, ysc ;  
  int i, kmark=2 ;

  SRAND(77) ;
  yd = 0.0 ;
  for (i=0; i<N; i++) {  
   p0[i] = DRAND()  ;
   p1[i] = DRAND()  ;
   if (i != kmark) p0[i] *= 0.1 ;
   if (i != kmark) p1[i] *= 0.1 ;
   yd = dis[i] = yd + 0.01*ranexp() ;
  }
  ysc = mutxhmmd(p0, p1, dis, N, kmark, theta, lambda) ;
  printf("theta; %9.3f\n", theta) ;
  printf("lambda; %9.3f\n", lambda) ;
  printf("score; %12.6f\n", ysc) ;
}
double
mutxhmmd(double *p0, double *p1, double *dis, int n, int kmark, double theta, double lambda) 
{
   double *x0, *x1, y1, y2 ;
  

   ZALLOC(x0, n, double) ;
   ZALLOC(x1, n, double) ;
   copyarr(p0, x0, n) ;
   copyarr(p1, x1, n) ;
   y1 = mutxhmm(x0, x1, dis, n, kmark, theta, lambda) ;
   x0[kmark] = x1[kmark] = 0.5 ;
   y2 = mutxhmm(x0, x1, dis, n, kmark, theta, lambda) ;
   free(x0) ;  free(x1) ;
   return y1-y2 ;
}
double
mutxhmm(double *p0, double *p1, double *dis, int n, int kmark, double theta, double lambda) 
{
   double *x0, *x1, *d0 , y1, y2 ;
   int km, l, r, k, i, e, kk  ;
   double yy, poisson, glen, temp  ; 
   double ***MM, **M ;
   double a0[3], a1[3], X[3] ;
   double b0[3], b1[3], ee[3] ;
   int x, y, numstates=3 ;
   int *lgtype, lpat, lmaxind ; 
   int *rgtype, rpat, rmaxind ; 
   double **L, **R ;
   double ylcheck, yrcheck, ycheck ; 
   double rcc[9] ;
   double yx, yscore, tt ;
  
   l = (n+1)/2 ;
   r = n-l ;
   if (kmark>=n) fatalx("kmark out of bounds\n") ;
   if (kmark< 0) fatalx("kmark out of bounds\n") ;
   if (kmark>=l) { 
    ZALLOC(x0, n, double) ;
    ZALLOC(x1, n, double) ;
    ZALLOC(d0, n, double) ;
    km = n-1-kmark ;
    if (km>=l) fatalx("logic bug\n") ;
    flipx(x0, p0, n) ;
    flipx(x1, p1, n) ;
    flipx(d0, dis, n) ;
    vst(d0, d0, -1.0, n) ;
    yy = mutxhmm(x0, x1, d0, n, km, theta, lambda) ;

    free(x0) ;
    free(x1) ;
    free(d0) ;
    return yy ;
   }


    ZALLOC(MM, n, double **) ;
    for (k=0; k<(n-1); ++k) { 
     M = MM[k] = initarray_2Ddouble(numstates, numstates, 0.0) ;
     poisson = 0.0 ;
     glen = dis[k+1]-dis[k] ;
/* this is in morgans */ 
     temp = lambda * glen ;
     if (temp<20.0) {
      poisson = exp(-temp) ;
     }
     calcM(poisson, theta, M, numstates) ;
    }

    lmaxind = pow((double) numstates, (double) l) ;
    ZALLOC(lgtype, l, int) ;
    rmaxind = pow((double) numstates, (double) r) ;
    ZALLOC(rgtype, r, int) ;

     L  = initarray_2Ddouble(numstates*lmaxind, numstates, 0.0) ;
     R  = initarray_2Ddouble(rmaxind, numstates, 0.0) ;

    for (lpat = 0; lpat < lmaxind; ++lpat) {  
     decode(lgtype, lpat, l, numstates) ;
     for (e = 0; e<numstates; ++e) {  
/* alphapass */
      sett1(a0, theta, numstates) ;
      k=0;
      calcx(X, p0[k], p1[k], lgtype[k], numstates) ;
      vvt(a1, a0, X, numstates) ;
      copyarr(a1, a0, numstates) ;
      if (k==kmark) { 
       vzero(a0, numstates) ;
       a0[e] = a1[e]; 
      }
      
    /* create alpha_star vector by matrix multiplication */
      for (k=1; k<l; ++k) {
       M = MM[k-1] ;
       calcx(X, p0[k], p1[k], lgtype[k], numstates) ;
       vzero(a1, numstates) ;
       for (x=0; x<numstates; x++) {
	for (y=0; y<numstates; y++) {
	  a1[x] += M[x][y] * a0[y];
/* x is new state y is old */
	}
       }
       vvt(a1,a1, X, numstates) ;
       copyarr(a1, a0, numstates) ;
       if (k==kmark) { 
        vzero(a0, numstates) ;
        a0[e] = a1[e]; 
       }
      }
      copyarr(a0, L[numstates*lpat+e], numstates) ;
     } // end e loop 
    }

 if (verbose) {
  ylcheck = 0 ;
  for (lpat=0; lpat<lmaxind; ++lpat) { 
   for (e=0; e<numstates; ++e) {  
    ylcheck += asum(L[numstates*lpat+e], numstates) ;
   }
  }
  printf("##lcheck1: %12.6f\n", ylcheck) ;
 }
 

    vclear(b0, 1.0, numstates) ;
    for (rpat = 0; rpat < rmaxind; ++rpat) {  
     vclear(b0, 1.0, numstates) ;
     decode(rgtype, rpat, r, numstates) ;
       for (k=r-1; k>=0; --k) { 
        kk = k+l ;
        M= MM[kk-1] ;
        calcx(X, p0[kk], p1[kk], rgtype[k], numstates) ;
        vvt(b1, b0, X, numstates) ;
        vzero(b0, numstates) ;
        for (x=0; x<numstates; x++) {
    	 for (y=0; y<numstates; y++) {
	  b0[x] += M[y][x] * b1[y];
         }
	}
       }
       copyarr(b0, R[rpat], numstates) ;
       if (verbose) {
        printf("rpat: %4d ", rpat) ;
        printmat(b0, 1, numstates) ;
       }
      }
/**  
 now put L, R together 
*/
 ylcheck = 0 ;
 for (lpat=0; lpat<lmaxind; ++lpat) { 
   for (e=0; e<numstates; ++e) {  
    ylcheck += asum(L[numstates*lpat+e], numstates) ;
   }
 }
 if (verbose) 
  printf("##lcheck2: %12.6f\n", ylcheck) ;
 ylcheck = yrcheck = ycheck = 0.0 ;
 vzero(rcc, 9) ;
 yscore = 0.0 ;
      
 if (verbose) {
  for (rpat=0; rpat<rmaxind; ++rpat) { 
   yrcheck += asum(R[rpat], numstates) ;
   vvp(rcc, rcc, R[rpat], numstates) ;
  }
  yrcheck = asum(rcc, numstates) ;
  printf("##rcc: ") ;
  printmat(rcc, 1, numstates) ;
 }

 sett1(a0, theta, numstates) ;
 for (lpat=0; lpat<lmaxind; ++lpat) { 
  for (rpat = 0; rpat <numstates; ++rpat) {
    for (e=0; e<numstates; ++e) {  
     if (rpat==0) {
      ylcheck += asum(L[numstates*lpat+e], numstates) ;
     }
     ee[e] = vdot(L[numstates*lpat+e], R[rpat], numstates) ;
    }
    y1 = asum(ee, numstates) ;  // P(g) 
    ycheck += y1 ;
    if (y1<=0.0) continue ;
    vst (b1, ee, 1.0/y1, numstates) ; // b1 scratch = P(e|g) 
    vvd (b0, b1, a0, numstates) ;     // P(e|g)/P(e) 
    tt = vldot(b1, b0, numstates) ;
    yscore += y1*tt ;
  }
 }

  if (verbose) 
   printf("## checks: %12.6f  %12.6f  %12.6f\n", ylcheck, yrcheck, ycheck) ;

 
    free(lgtype) ;
    free(rgtype) ;

    for (k=0; k<(n-1); ++k) { 
      free2D(&MM[k], 3) ; 
    }
    free2D(&L, numstates*lmaxind) ;
    free2D(&R, rmaxind) ;
    free(MM) ;
    return yscore ;
}
void flipx (double *a, double *b, int n) 
{
/* a, b must not overlap */
     int i ;

     for (i=0; i<n; i++)  {  
      a[i] = b[n-i-1] ;
     }
}

void calcM(double poisson, double theta, double **M, int numstates)
{
  double PrCC, PrAA;

  PrCC = poisson + (1-poisson)*theta;	
  PrAA = poisson + (1-poisson)*(1-theta);
  
  if (numstates == 3) {
    M[0][0] = PrAA * PrAA;
    M[1][0] = 2 * PrAA * (1-PrAA);
    M[2][0] = (1-PrAA) * (1-PrAA);
    
    M[0][1] = PrAA * (1-PrCC);
    M[1][1] = (1-PrAA) * (1-PrCC) + PrAA * PrCC;
    M[2][1] = (1-PrAA) * PrCC;
    
    M[0][2] = (1-PrCC) * (1-PrCC);
    M[1][2] = 2 * PrCC * (1-PrCC);
    M[2][2] = PrCC * PrCC;
  }
  else if (numstates == 2) {
    M[0][0] = PrAA;
    M[1][0] = (1-PrAA);
    M[0][1] = (1-PrCC);
    M[1][1] = PrCC;
  }

}

void calcx(double *X, double PopA_freq, double PopB_freq, int gtype, int numstates) 
{
  
   if (numstates==2) {

    switch(gtype) {

     case 1:
      X[0] = PopA_freq;
      X[1] = PopB_freq;
      break;

     case 0:
      X[0] = (1 - PopA_freq);
      X[1] = (1 - PopB_freq);
      break;

     default:  
      fatalx("bad gtype %d\n", gtype) ;

    }
    return ;
   }
  
  switch (gtype)  {
      
    case 2:
      X[0] = PopA_freq * PopA_freq;
      X[1] = PopA_freq * PopB_freq;
      X[2] = PopB_freq * PopB_freq;
      break;
      
    case 1:
      X[0] = 2 * PopA_freq * (1 - PopA_freq) ;
      X[1] = PopB_freq * (1 - PopA_freq) + 
	(1 - PopB_freq) * PopA_freq ;
      X[2] =  2 * PopB_freq * (1 - PopB_freq) ;
      break;
      
    case 0:
      X[0] = (1 - PopA_freq) * (1 - PopA_freq);
      X[1] = (1 - PopA_freq) * (1 - PopB_freq);
      X[2] = (1 - PopB_freq) * (1 - PopB_freq);
      break;

     default:  
      fatalx("bad gtype %d\n", gtype) ;
  }
  
}
void
decode(int *gtype, int pat, int l, int base)       
{
 int tpat ;
 if (l==0) return ;
 if (l==1) { 
   gtype[0] = pat ;
 }
 gtype[l-1] = pat % base ;
 tpat = pat / base ;
 decode(gtype, tpat, l-1, base) ;
}
