#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <nicklib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "tlest.h"
#include "sampgw.h"
#include "arms.h"


extern int verbose ;
extern int *fchrom, *lchrom ;
extern SNP **snpmarkers ;

extern double muval1  ;
extern double thp1, thp2, thxp1, thxp2 ;  
extern double lp1, lp2, lxp1, lxp2 ;  
extern double thxp0 ;
extern int rdismode ;
extern double rdis ;

double sampit1(double p, double lam, 
 double lo, double hi, double roll) ;

void sampprbt(double *theta, double xtheta, 
 double ta, double tb, double c, double a, double b, double qval)  ;

void
sampzld(double *pz, double xp1, double xp2, 
  double *xdis, double *pa, double *lama, int n)  ;

double gproba(double *x, double *p, double *lam, int n)  ;

void
tlest(Indiv *indx, SNP **snpmarkers, int *ethnic, int fc, int lc,
   double *theta, double *lambda)  
{
   SNP *cupt, *cupt2 ;
   int clo, chi, tfc, tlc ;
   int numstates, ignore, numx, chrom ; 
   int k, k1, k2, kk2, a, b, aa[2], bb[2] ;
   int kpoiss, tkpoiss, giter = 0 ; 
   double ttheta, tlambda ; 
   double tt[2], tuu[2], utot[2], stot[2] ; 
   double ldis, dis, dl ;
   double ldistot ;
   double tp, tl ; /* shape parameter for posterior on lambda */
   static int ncall=0 ;
   int  nhidden, nh1, nh0, tkcheck ;
   double q1, q2, r1, r2 ;
   double atheta, xtheta ;
   double mu ;

/* 
   number of events not observable because a second one occurs
   before a real marker 
*/

   ++ncall ;

   atheta = indx -> theta_mode ;
   xtheta = indx -> Xtheta_mode ;

   cupt = snpmarkers[fc] ;
   gettln(cupt, indx, &ttheta, &tlambda, &numstates, &ignore) ;



   if (ttheta==0) fatalx("ttheta zero\n") ;
   clo = cupt -> chrom ;

   if (ignore == YES)  { 
/** no data */
    if (clo != 23) {  
     q1 = thp1 ;
     q2 = thp2 ;
     r1 = lp1 ; r2 = lp2 ;
    }
    else {  
     q1 =  thxp1 ;
     q2 =  thxp2 ;
     q1 += thxp0 * atheta ;
     q2 += thxp0 * (1.0-atheta) ;
     r1 = lxp1; r2 = lxp2 ;
    }
    ttheta  = ranbeta(q1, q2) ;
    tlambda = rangam(r1)/r2 ;
    *theta  = ttheta ;
    *lambda = tlambda ;
    return ;
   }



   numx = numstates -1 ;

   clo = cupt -> chrom ;
   cupt = snpmarkers[lc] ;
   chi = cupt -> chrom ;

   vzero(utot,2) ;
   vzero(stot,2) ;
/** 
 stot counts events at start of chromosome 
*/
   ldistot = 0.0 ;
   kpoiss = 0 ;

   for (chrom = clo; chrom <= chi;  ++chrom)  {
    if (lchrom[chrom] < 0) continue ;
    tfc = MAX(fchrom[chrom], fc) ;
    tlc = MIN(lchrom[chrom], lc) ;
    k1 = nextmarkx(tfc, tlc, 1, ethnic) ;
    if (k1<0) continue ;
    cupt = snpmarkers[k1] ;
    a = ethnic[k1] ;
    genaa(a, aa, numstates) ;

    for (k=0; k<numx; k++) {  
     b = aa[k] ;
     if (b>=0) ++stot[b] ;
    }

    for (;;)  {  
     cupt = snpmarkers[k1] ;
     cupt2 = (SNP *) cupt -> pcupt ;
     if (cupt2 == NULL) break ; 
     k1 = cupt -> markernum ;
     k2 = cupt2 -> markernum ;
     kk2 = nextmarkx(k1+1, tlc, 1, ethnic) ;
     dis = cupt->dis ;
     a = ethnic[k1] ;
     b = ethnic[k2] ;
     if (ncall==-1) {
      printf ("zz %15s %15s", cupt -> ID, cupt2 -> ID) ;
      printf(" %12.6f %3d %3d %4d %4d %4d\n", dis, a, b, k1, k2, kk2) ;
     }
     ldis = ((double) numx) * dis ;
     dl = dis*tlambda ;
     psim2(dl, ttheta, a, b, tuu, &tkpoiss, numstates) ;
     ldistot += ldis ;
     vvp(utot, utot, tuu, 2) ;
     kpoiss += tkpoiss ;
     cupt -> esum += tkpoiss ;
     k1 = k2 ;
    }
   }

/** 
 now look at hidden events 
 and generate appropriate number of 
 PopB blocks 
*/

/**
   We regard theta, lambda as generating an "event list"  
   # such events | >1 in interval is independent of theta.  
   Therefore it is valid to consider just the last event
   in each gap
*/

   vvp(utot, utot, stot, 2) ;

   q1 = utot[1] ;
   q2 = utot[0] ;

   if (clo != 23) {  
    q1 += thp1 ;
    q2 += thp2 ;
   }
   else {  
    q1 += thxp1 ;
    q2 += thxp2 ;
   }


   if (thxp0 <0) {  
/* old style */
    ttheta = ranbeta(q1, q2) ;
   }

   else {  
    if (clo == 23) {  
     q1 += thxp0 * atheta ;
     q2 += thxp0 * (1.0-atheta) ;
     sampbmu(&ttheta, q1, q2, muval1) ;
    }
    if (clo != 23) {  
     sampprbt(&ttheta, xtheta, q1, q2, 
      thxp0, thxp1, thxp2, indx -> qval) ;
    }
/* qval has no effect unless qtmode */
   }

    q1 = lp1 ; q2 = lp2 ;
    if (clo == 23) {
     q1 = lxp1; q2 = lxp2 ;
    }

    tp = q1 + (double) kpoiss  ;
    tl = ldistot + q2 ;

    tlambda = rangam(tp)/tl ;

    *theta  = ttheta ;
    *lambda = tlambda ;


/** now add up lsum with new lambda value */  
   for (chrom = clo; chrom <= chi;  ++chrom)  {
    if (lchrom[chrom] < 0) continue ;
    tfc = MAX(fchrom[chrom], fc) ;
    tlc = MIN(lchrom[chrom], lc) ;
    k1 = nextmarkx(tfc, tlc, 1, ethnic) ;
    if (k1<0) continue ;

    for (;;)  {  
     cupt = snpmarkers[k1] ;
     cupt2 = (SNP *) cupt -> pcupt ;
     if (cupt2 == NULL) break ; 
     k2 = cupt2 -> markernum ;
     cupt -> lsum += ((double) numx) * tlambda  ;
     k1 = k2 ;
    }
   }


   return ;
}

void
genaa(int a, int *aa, int numstates)  
{
  int xx[2] ; 
  genab(0.0, 0.0, 0, a, 
    xx, aa, numstates)  ;
}

void
genab(double dl, double ttheta, int a, int b, 
 int *aa, int *bb, int numstates) 
/**
 a, b are number of PopB ancestries 
 return pair (A0, B0),  (A1, B1) of ancestries on chromosome.  
 case a=b=1 there are 2 possibilities and we randomly choose
*/

{
   int a0, b0, a1, b1 ;
   double t, x0, x1 ;
   double p, y0, y1 ;
   double tt[2] ;
   
   if ((a < 0) || (a>2)) fatalx("bad ethnic %d\n,a") ;
   if ((b < 0) || (b>2)) fatalx("bad ethnic %d\n",b) ;

   if (numstates==2) {  
    if (a>=2) fatalx("bad ethnic (2)  %d\n",a) ;
    if (b>=2) fatalx("bad ethnic (2)  %d\n",b) ;
    aa[0] = a ;
    bb[0] = b ;
    aa[1] = bb[1] = -1 ;
    return ;
   }

   a0 = -1 ;
   if (a != 1) {  
    a0 = a1 = a/2 ;
    b0 = b/2 ;
    b1 = b - b0 ;
   }

   if (b != 1) {  
    b0 = b1 = b/2 ;
    a0 = a/2 ;
    a1 = a - a0 ;
   }

   if (a0<0) {  

/** 
 a=b=1
*/

    a0=0; a1=1 ;
    tt[0] = 1-ttheta ;
    tt[1] = ttheta ;
    t = exp(-dl) ;
    x0 = t + (1-t)*tt[0] ;
    x1 = t + (1-t)*tt[1] ;
    y0 = (1-t)*tt[1] ;
    y1 = (1-t)*tt[0] ;
    p = (x0*x1) / (x0*x1+y0*y1) ;
    if (DRAND()<p)  {
      b0 = 0; 
    }
    else  {
      b0 = 1; 
    }
    b1 = 1-b0 ;
   }

   aa[0] = a0 ;
   aa[1] = a1 ;

   bb[0] = b0 ;
   bb[1] = b1 ;

   return ;
}

void psim1
 (double dl, double theta, int a, int b, 
   double *uu, int *kpoiss) 
{
  double th, p0, p1, y, t ;  
  vzero(uu,2) ;
  *kpoiss = 0; 
  if (b>2) fatalx("(psim1) large b: %d\n",b) ;
  if (a==b) {  
/* was there a hidden event */
   t = exp(-dl) ;
   th = theta ;
   if (b==0) th = 1.0-theta ;
   p0 = t ;  
   p1 = (1-t)*th ;
   y = p0/(p0+p1) ; /* prob no event */
   if (DRAND() < y) return ;
  }
  *kpoiss = (int) ranpoissx(dl) ; /*x =  poisson |  x > 0  */
  ++ uu[b] ;
}

void psim2
 (double dl, double theta, int a, int b, 
   double *uu, int *kpoiss, int numstates) 
{
  double u1[2], u2[2] ;
  int aa[2], bb[2], k1, k2 ;
  static int ncall=0 ;

  ++ncall ;

  if ((a<0) || (b<0))  
   fatalx("(psim2) bad ethnic: %d %d\n", a, b) ;

  genab(dl, theta, a, b, aa, bb, numstates) ;
  k2 = 0 ;
  vzero(u2, 2) ;
  psim1(dl, theta, aa[0], bb[0], u1, &k1) ;
  if (numstates == 3) {
   psim1(dl, theta, aa[1], bb[1], u2, &k2) ;
  }
  vvp(uu, u1, u2, 2); 
  *kpoiss = k1+k2 ;
}

int nextmarkx (int start, int end, int inc, int *ee) 
/* search for real marker with ee[] >=0  */
{
  int x ;
  SNP *cupt ;
  x = start ;
  if ( (x<end) && (inc <0)) 
    fatalx("bad nextmarkx %d %d %d\n",start, end, inc ) ;
  if ( (x>end) && (inc >0)) 
    fatalx("bad nextmarkx %d %d %d\n",start, end, inc ) ;
  for (x=start; x != end ; x += inc)  {
   cupt = snpmarkers[x] ;
   if (!(cupt -> isfake) && !(cupt ->ignore) && (ee[x] >=0))  return x ;
  }
  return -1 ;
}
void setestdis(SNP **snpm, int nsnps) {  

 int k ; 
 SNP *cupt1, *cupt2 ;

 clearestdis(snpm, nsnps) ;
 
 for (k=0; k< nsnps ; ++k)  { 
  cupt1 = snpm[k] ;
  if (cupt1 -> ignore) continue ;
  if (cupt1 -> isfake) continue ;
  cupt2  = (SNP *) cupt1 -> pcupt ;
  if (cupt2 == NULL) {  
   cupt1 -> estdis = 1000.0 ;
  }
  else  { 
   cupt1 -> estdis = cupt2 -> genpos - cupt1 -> genpos ;  
  }
   cupt1 -> dis = cupt1 -> estdis ;  
 }
}

void cleargpsum(SNP **snpm, int nsnps) { 
 SNP *cupt ; 
 int k ;

 for (k=0; k< nsnps ; ++k)  { 
  cupt = snpm[k] ;
  cupt -> gpsum  = cupt -> gpnum = 0 ; 
 }
}
void setgpsum(SNP **snpm, int nsnps) { 
 SNP *cupt ; 
 int k ;

 for (k=0; k< nsnps ; ++k)  { 
  cupt = snpm[k] ;
  cupt -> gpsum += cupt -> genpos ; 
  cupt -> gpnum += 1  ; 
 }
}

void clearestdis(SNP **snpm, int nsnps) { 

 int k, chrom, lchrom ; 
 SNP *cupt, *ncupt ;
 
 ncupt = (SNP *) NULL ;
 lchrom = -1 ;

 for (k=nsnps-1; k>=0  ; --k)  { 
  cupt = snpm[k] ;
  cupt -> esum = cupt -> lsum = 0 ;
  if (cupt -> ignore) continue ;
  if (cupt -> chrom != lchrom)  { 
   ncupt = NULL ;
  }
  lchrom = cupt -> chrom ;  
  cupt -> pcupt = ncupt ;  
  if (cupt -> isfake == NO)   ncupt = cupt ;
 }
}

void reestld(SNP **snpm, int nsnps, Indiv **indm, int numind, double xflat) 
{
/** ray sampling to simultaneously estimate lam, dis  */  
     double *xdis, *priordis ; 
     double *pa, *lama ;
     SNP **slist, *cupt ;
     Indiv *indx ;
     int n, lo, hi, i, k ; 
     double z, zold, dis ;
     double xp1, xp2;

     if (rdismode == NO) return;

     ZALLOC(priordis, nsnps, double) ;
     ZALLOC(xdis, nsnps, double) ;
     ZALLOC(slist, nsnps, SNP *) ;


     lo = fchrom[1] ;
     hi = lchrom[22] ;
     n = loadpdis(slist, priordis, xdis, snpmarkers, lo, hi) ;
     ZALLOC(pa, n, double) ;
     ZALLOC(lama, n, double) ;
     vst(pa, priordis, xflat, n) ;
     vclear(lama,  xflat, n) ;

     zold = z = ranexp() ; 
     vst(xdis, xdis, 1.0/z, n) ;  
     xp1 = lp1 ; 
     xp2 = lp2/z ;
     sampzld(&z, xp1, xp2, xdis, pa, lama, n) ;
     
     vst(xdis, xdis, z, n)  ;
     lp2 *= (z/zold) ;
     for (k=0; k<n; k++)  { 
      cupt = slist[k] ;
      dis = cupt -> dis *= (z/zold) ;
      fixgpos(snpm, cupt, dis) ;
     }
     for (i=0; i<numind; i++)  { 
      indx = indm[i] ; 
      indx -> lambda_mode /= (z/zold) ;
     }
    
     free(pa) ; 
     free(lama) ;

     lo = fchrom[23] ;
     hi = lchrom[23] ;
     n = loadpdis(slist, priordis, xdis, snpmarkers, lo, hi) ;
     ZALLOC(pa, n, double) ;
     ZALLOC(lama, n, double) ;
     vst(pa, priordis, xflat, n) ;
     vclear(lama,  xflat, n) ;

     zold = z = ranexp() ; 
     vst(xdis, xdis, 1.0/z, n) ;  
     xp1 = lxp1 ; 
     xp2 = lxp2/z ;
     sampzld(&z, xp1, xp2, xdis, pa, lama, n) ;
     
     vst(xdis, xdis, z, n)  ;
     lxp2 *= (z/zold) ;
     for (k=0; k<n; k++)  { 
      cupt = slist[k] ;
      dis = cupt -> dis *= (z/zold) ;
      fixgpos(snpm, cupt, dis) ;
     }
     for (i=0; i<numind; i++)  { 
      indx = indm[i] ; 
      indx -> Xlambda_mode /= (z/zold) ;
     }
    
     free(pa) ; 
     free(lama) ;


     free(priordis) ; 
     free(xdis) ; 
     free(slist) ;


}


void reestdis(SNP **snpm, int nsnps, Indiv **indm, int numind, double xflat) 
{ 

// estimate genetic map.  Likely to mix quite poorly 

 int k, j, jmax ; 
 SNP *cupt1, *cupt2, *cupt ;
 double ta, tl, basepos, y, dis, estdis ; 
 static int ncall = 0 ;
 static int newmode = NO  ; 

 if (rdismode == NO) return ;
 ++ncall ;
 for (k=0; k< nsnps ; ++k)  { 
  cupt1 = snpm[k] ;
  if (cupt1 -> ignore) continue ;
  if (cupt1 -> isfake) continue ;
  cupt2  = (SNP *) cupt1 -> pcupt ;
  if (cupt2 == NULL) continue ;
   estdis = cupt1 -> estdis ; 

  if (newmode == YES) {
   dis = dodis(cupt1, cupt2, indm, numind, xflat) ;
   cupt1 -> dis = dis ; 
   if (ncall < 50) 
   printf("zzff %s %12.6f %12.6f %12.6f\n", cupt1 -> ID, dis, cupt1 -> estdis, xflat) ;
  }

  if (newmode == NO) {
   ta = estdis*xflat ;  
   tl = xflat ;   
   ta += cupt1 -> esum ;  
   tl += cupt1 -> lsum ;  
   
   cupt1 -> dis = dis = rangam(ta)/tl ;
  }


   if (verbose) 
    printf("zzzdis %9.3f %9.3f %9.3f %9.3f\n", dis, estdis, cupt1 -> esum, cupt1 -> lsum) ;

   if (dis>10.0) fatalx("absurd distance\n") ;
   fixgpos(snpm, cupt1, dis) ;
 }
}
void fixgpos(SNP **snpm, SNP *cupt1, double dis) 
{

   double basepos ;
   int j, k, jmax ;
   SNP *cupt, *cupt2  ;
   double  y, estdis, oldpos ;

   cupt2 = cupt1 -> pcupt ;
   if (cupt == NULL) return ;
   if (dis>10.0) fatalx("absurd distance\n") ;

   estdis = cupt1 -> estdis ;
   basepos = cupt1 -> genpos ;
   k = cupt1 -> markernum ;
   jmax = cupt2 -> markernum ;
   for (j=k; j <= jmax; ++j) {  
    cupt = snpm[j] ;
    y = (cupt -> estgenpos - cupt1 -> estgenpos) / estdis ;
    oldpos = cupt -> genpos ;
    cupt -> genpos = basepos + y * dis ;
    // printf("zzpos %12.6f  %12.6f  %12.6f\n", oldpos, cupt -> genpos, dis) ; 
   }


}
int 
loadpdis(SNP **slist, double *priordis, double *xdis, 
   SNP **snpm,  int lo,  int hi) 
{

     SNP *cupt, *cupt2 ; 
     int i, n=0 ;
     double t ;

     for (i=lo; i<=hi; i++)  {  
    
      cupt = snpm[i] ;

      if (cupt -> isfake) continue ;
      if (cupt -> ignore) continue ; 
      cupt2 = cupt -> pcupt;      
      if (cupt2 == NULL) continue ;
      t = priordis[n] = cupt -> estdis ;
      if (t<0.0) fatalx("zzbad %d %s %12.6f  %12.6f\n", 
       i, cupt -> ID, cupt -> estgenpos, cupt -> estdis) ;

      if (t==0.0) continue ;

      xdis[n] = cupt -> dis ;
      if (slist != NULL) slist[n] = cupt ;
      ++n ;
     }
     return n ;
}

void estflat (double *flat, SNP **snpm, int nsnps) 
{
     double *priordis, *xdis ;
     SNP *cupt, *cupt2 ;
     int i, j, n=0 ;
     double tflat ;
     static int ncall = 0 ;

     ++ncall ;

     ZALLOC(priordis, nsnps, double) ;
     ZALLOC(xdis, nsnps, double) ;

     n = loadpdis(NULL, priordis, xdis, snpmarkers, 0, nsnps-1) ;

     tflat = *flat ;
     samppdis(&tflat, xdis, priordis, n, 10.0, 10000.0) ;
     *flat = tflat ;

     free(priordis) ;
     free(xdis) ;
}

double
dodis(SNP *cupt1, SNP *cupt2, Indiv **indm, int numind, double xflat)
{

  double dis, ylike, ybase, initdis ; 
  int k ;
  struct xparm xdata, *dpt ;
  double xinit [8] ;
  int ninit1, ninit2 ;

  double xsamp[10], xcent[10], qcent[10] = {5., 30., 70., 95.};
  double convex = 1.;
  double xl, xr, xprev ;
  double xmul = 1.5 ;
  int dometrop = 1;
  int mnum ;
  
  int err, ninit = 3, npoint = 101, nsamp = 1, ncent = 0 ;
  int neval;

  dpt = &xdata ;  
  dpt -> cupt1 = cupt1 ;
  dpt -> cupt2 = cupt2 ;
  dpt -> indm =  indm ;
  dpt -> numindivs = numind ;
  dpt -> xflat = xflat ;
  initdis = cupt1 -> estdis ;

  dis = initdis/(xmul*xmul) ;
  mnum = cupt1 -> markernum ;

  ninit = 5 ;
  for (k=1; k<=ninit; ++k) {  
   xinit[k-1] = dis ;
   ylike = ldis(cupt1, cupt2, indm,  numind, xflat, dis)  ;
   if ((verbose) || (mnum==0))   {
    printf("dis: %12.6f  llike: %12.6f\n", dis,  ylike) ;
   }
   dis *= xmul ;
  }

   xl = 1.0e-8 ;
   xr = MAX(30.0, dis+1) ;
   xr = MAX(xr, initdis+1) ;
   xprev = initdis ; 
   err = arms(xinit,ninit,&xl,&xr,xldis, &xdata, &convex,
           npoint,dometrop,&xprev,xsamp,nsamp,
           qcent,xcent,ncent,&neval);

   if (err>0) {
    printf("errval:  %d\n", err) ;
    fatalx("bad arms\n") ;

   }

   dis = xsamp[0] ;
   return dis ;
}

double xldis(double x, void *xdata) 
{
  struct xparm *d;
  double y ;

  d = xdata ; 
  y =  ldis(d -> cupt1, d -> cupt2, d -> indm, d -> numindivs, d -> xflat, x) ;
  if (verbose)
   printf("zzz %12.6f %12.6f\n", x, y) ;

  return y ; 

}
double xldis2(double x, void *xdata) 
{
  struct xparm2 *d;
  double yl, z, priorl, yl1, yl2 ;
  double *xx ;  
  double yp1, yp2, mean, var ;
  int n ;

  d = xdata ; 
  n = d->n ;
  z = x ;
  yp1 = d -> xp1 ;
  yp2 = d -> xp2 * z ;
  mean = yp1/yp2 ;
  var =  mean/yp2 ;
  ZALLOC(xx, n, double) ;
  vst(xx, d -> xdis , z, n) ;

  yl1 = gproba(xx, d -> pa, d -> lama, n) ;
  yl2 = (double) (n+1) * log(z) ; 
  yl2 -= z ;

  yl = yl1 + yl2 ;

  priorl = priorgl(mean, var) ;

  yl += priorl ;  

  free(xx) ;
  if (verbose) 
   printf("zz2 %12.6f %12.6f %12.6f %12.6f %12.6f\n", x, yl1, yl2, priorl,yl) ;
  return yl  ; 

}
double 
ldis(SNP *cupt1, SNP *cupt2, Indiv **indm, int numind, double xflat, double dis) 
{
  int k, numstates = 3, ignore ;
  Indiv *indx ;
  double yscore= 0.0, yl, poisson, y1   ;
  double **M ;  
  double temp, theta, lambda,  ta, tb ;          
  int x, y ;

  M = initarray_2Ddouble(3, 3, 0.0) ; 
  yl = 0.0 ;
  ta = cupt1 -> estdis  * xflat ;
  ta = MIN(ta, 1.0e-8) ;
  tb = xflat ;
  yl = (ta-1.0)*log(dis) - tb*dis ;
  for (k=0; k< numind; ++k)  { 
    indx = indm[k] ;
    gettln(cupt1, indx, &theta, &lambda, &numstates, &ignore) ;
    if (ignore) continue ;
     y = getep(cupt1, k) ;
     x = getep(cupt2, k) ;
     if (y<0) fatalx("bad ep %s\n", cupt1 -> ID) ;
     if (x<0) fatalx("bad ep %s\n", cupt2 -> ID) ;
     temp = lambda * dis ;
     if (temp<20.0) {
      poisson = exp(-temp) ;
     }
     else poisson = 0.0 ;
     if (lambda>500) poisson = 0.0 ;
     calcM(poisson, theta, M, numstates) ;
     yl += log(M[x][y]) ;
     if (cupt1 -> markernum == 0) printf("zzk %d %d\n", x, y) ; 
     if (cupt1 -> markernum == 5632) printf("zzkx %d %d\n", x, y) ; 
/* x new state, y old */
  }
  free2D(&M, 3) ;
  return yl ;
}

void
sampzld(double *pz, double xp1, double xp2, 
  double *xdis, double *pa, double *lama, int n) 
{

   double z, zold, t, tt ; 
   int i ;
   double yl, priorl, yl1, yl2, yl3 ;
   double mean, var ;   

  struct xparm2 xdata, *dpt  ;
  double xinit [8] ;
  int ninit1, ninit2 ;
  double xsamp[10], xcent[10], qcent[10] = {5., 30., 70., 95.};
  double convex = 1.;
  double xl, xr, xprev ;
  double xmul = 1.1 ;
  int dometrop = 1;
  
  int err, ninit = 3, npoint = 101, nsamp = 1, ncent = 0 ;
  int neval;




   dpt = &xdata ;
   dpt -> xdis = xdis ;
   dpt -> pa = pa ; 
   dpt -> lama = lama ;
   dpt -> n = n ;
   dpt -> xp1 = xp1 ;
   dpt -> xp2 = xp2 ;

   zold = z = *pz ;

   xinit[0] = z/xmul  ;
   xinit[1] = z  ;
   xinit[2] = z*xmul  ;

//   printf("z: %9.3f\n", z) ;

   xl = 1.0e-8 ;
   xl = MIN (xl, 0.99*xinit[0]) ;
   xr = 30.0 ;
   xr = MAX (xr, 1.01*xinit[2]) ;

   xprev = z ; 
   err = arms(xinit,ninit,&xl,&xr,xldis2, &xdata, &convex,
           npoint,dometrop,&xprev,xsamp,nsamp,
           qcent,xcent,ncent,&neval);

   if (err>0) fatalx("bad sampz %d\n", err) ;

   z = *pz = xsamp[0] ;  

//   printf("znew: %9.3f  rat: %9.3f\n", z, z/zold) ;
    


}
double gproba(double *x, double *p, double *lam, int n) 
{
   int i ; 
   double yl = 0.0 ;

   for (i=0; i<n; i++) {  
     yl += gammprob(x[i], p[i], lam[i]) ;
   }
   return yl ;
}

