#include "toysim.h" 
#include "qsubs.h" 
#include <math.h>
#include "gammas.h" 


double t1val = 0.5, txval = 0.3 ;
double l1val = 8.0, lxval = 8.0 ;

extern double a1, b1, p1, psi1 ;
double aa2 = 2, bb2= 14, cc2 = 85 ;

double px1 = 9, psix1 = 2 ;

static int debug = NO ;
static int xnum1 = -1, xnum2 = -1, enum1, enum2  ;

extern int alldata ;
extern int theta2mode ;  
extern int lambda2mode ;  
extern int details ;
extern double sim2dvals[4] ;
extern int sim2d_caseonly ;

extern int qtmode ;  
extern double *qtrho, *qtrisks ;  
extern double qtrbase ;

static double genth1(double a, double b) ;
static double genlam1(double a, double b, int mode) ;
static int xlmrisk(int a, int g, double lam, double mu, int numstates) ;

void 
dosim(SNP **snpmarkers, Indiv **indivmarkers,
 int numsnps, int numindivs, double xrisk, double xrho,
 int markersim, double *ttau) 
{
    int *ethnic, *gtypes, *ethnic2 ;
    Indiv *indx ;
    int indiv, numstates ; 
    int ignore ;
    double theta, lambda ; 
    double y, t, y2 ;
    double ya, yx ;
    double t0[3], ss[3], psi[3], gg[3],  g2[3] ;
    double theta1, theta2, risk ;
    double lambda1, lambda2, lam ;
    SNP *cupt ;
    int fc, lc, chrom, i, a, g, k, j, ghalf, kode ;
    static int ncall=0; 
    int n1, ntot, numx ;
    double *eauto, *ex ;
    double *tt, *ttx ;
    int mnum ;
    double xa, xb, p ;
    int noxdata = NO ;
    int sim2dmode = NO ;
    double rlam, rmu ;
    char *mname = NULL ;
    double qcc[3] ; 

    int nn[3] ;   

    monitlo = -1 ; monithi = -2 ;
    ivzero(nn, 3) ;

    risk = xrisk ;
    if (qtmode) { 
      setqcox(qcc, a1, b1, xrho, xrisk) ;
    }

    ++ncall ;
    printf("toysim called\n") ;
    if(markersim >= numsnps) fatalx("Marker number %d to be simulated out of bounds\n",markersim);
    if (markersim>=0) mname = snpmarkers[markersim] -> ID ;
    else mname = strdup ("") ;
    printf("markersim: %d name: %s risk: %9.3f\n", 
    markersim, mname, risk) ;
    printf("tauscal: %9.3f %9.3f\n", ttau[0], ttau[1]) ;
    printf("a1: %9.3f b1: %9.3f\n", a1, b1) ;
    printf("p1: %9.3f psi1: %9.3f\n", p1, psi1) ;

    if (sim2dvals[0] >= 0.0) {  
     sim2dmode = YES ;
     risk = 1.0 ;
     printf("sim2dmode\n") ;
    }

    for (i=0; i<numsnps; i++)  {  
     cupt = snpmarkers[i] ;
     if (cupt -> isfake) continue ;
     if (cupt -> ignore) continue ;  
    if (cupt -> popA_nn[0] >= 0) {  
     xa = cupt -> popA_nn[0]  ;
     xb = cupt -> popA_nn[1]  ;
     p = ranbeta(xa,xb) ;
    }
    else p = cupt->PopA_freq ;
    cupt -> PopAftrue = p ;
    gx(& (cupt -> adm_PopA_freq), p,  ttau[0]) ;
    cupt -> adm_PopAftrue = cupt -> adm_PopA_freq ;
    
    if (cupt -> popB_nn[0] >= 0) {  
     xa = cupt -> popB_nn[0]  ;
     xb = cupt -> popB_nn[1]  ;
     p = ranbeta(xa,xb) ;
    }
    else p = cupt->PopB_freq ;
    cupt -> PopBftrue = p ;
    gx(& (cupt -> adm_PopB_freq), p, ttau[1]) ;
    cupt -> adm_PopBftrue = cupt -> adm_PopB_freq ;
   }

   if (sim2dmode)  {  
// we don't worry here about modern and ancestral freqs being different
     if ((markersim <0) || (markersim>=numsnps)) 
       fatalx("sim2dmode.  risk marker not set\n") ; 
    cupt = snpmarkers[markersim] ;
    p = sim2dvals[0] ;
    cupt -> PopAftrue = cupt -> PopA_freq = p ;
    cupt -> adm_PopAftrue = cupt -> adm_PopA_freq = p ;
    p = sim2dvals[1] ;
    cupt -> PopBftrue = cupt -> PopB_freq = p ;
    cupt -> adm_PopBftrue = cupt -> adm_PopB_freq = p ;
    ivclear(cupt -> popA_nn, 1, 2) ;
    ivclear(cupt -> popB_nn, 1, 2) ;
   }

    ZALLOC(ethnic, numsnps, int) ;
    ZALLOC(ethnic2, numsnps, int) ;
    ZALLOC(gtypes, numsnps, int) ;
    ZALLOC(eauto, numindivs, double) ;
    ZALLOC(ex, numindivs, double) ;
    ZALLOC(tt, numindivs, double) ;
    ZALLOC(ttx, numindivs, double) ;

    a1 = MAX(a1, 1.0) ;
    aa2 = MAX(aa2, 1.0) ;
    b1 = MAX(b1, 1.0) ;
    bb2 = MAX(bb2, 1.0) ;

    printf("a1: %9.3f b1: %9.3f\n", a1, b1) ;
    printf("aa2: %9.3f bb2: %9.3f cc2: %9.3f\n", aa2, bb2, cc2) ;
    for (indiv = 0; indiv < numindivs; ++indiv) { 
     ivclear(ethnic, -1, numsnps) ;
     ivclear(ethnic2, -1, numsnps) ;
     ivclear(gtypes, -1, numsnps) ;

     indx = indivmarkers[indiv] ;

     theta1 = theta2 = t = ranbeta(a1,b1) ;
     lambda1 = lambda2 = lam = rangam(p1)/psi1 ;

     if (theta2mode == YES) {  
      theta1 = genth1(a1, b1) ;
      theta2 = genth1(a1, b1) ;
      t = 0.5*(theta1+theta2) ;
     }

     if (lambda2mode == YES) {  
      lambda1 = genlam1(p1, psi1, lambda2mode) ;
      lambda2 = genlam1(p1, psi1, lambda2mode) ;
      lam = 0.5*(lambda1+lambda2) ;
     }

     tt[indiv] =  indx -> theta_mode   =  t ; 
     ttx[indiv] = indx -> Xtheta_mode  = 
        ranbeta(cc2*t + aa2, cc2*(1-t) + bb2) ;

     indx -> thetatrue = tt[indiv] ;
     indx -> Xthetatrue = ttx[indiv] ;

     indx -> lambdatrue = indx -> lambda_mode   =  lam ; 
     indx -> Xlambdatrue = indx -> Xlambda_mode = rangam(px1)/psix1 ;

     for (chrom = 1; chrom <=23 ; ++chrom) { 

      fc = fchrom[chrom] ;
      lc = lchrom[chrom] ;
      if ( (fc > 99999) || (lc <0)) continue ;
      cupt = snpmarkers[fc] ;
      gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;


      mnum = -1 ;
      if (indx -> affstatus) mnum = markersim ;
 
      for (;;)  {
       if (chrom == 23) 
        ranpathr(theta, lambda, numstates, snpmarkers, ethnic, 
         fc, lc, mnum, risk) ;

       else {
        ranpathr(theta1, lambda1, 2, snpmarkers, ethnic, 
         fc, lc, mnum, risk) ;
        ranpathr(theta2, lambda2, 2, snpmarkers, ethnic2, 
         fc, lc, mnum, risk) ;
        ivvp(ethnic+fc, ethnic2+fc, ethnic+fc, lc-fc+1) ;
       }
       mkgtypes(gtypes, snpmarkers, ethnic, indiv, numstates, fc, lc) ;
// gtypes is array indexed by snpnumber.  Dangerous bend !!
/** 
 Note if affstatus = NO mnum == -1 and we break at next line 
*/
       if (mnum<fc) break ;
       if (mnum>lc) break ;
       a = ethnic[mnum] ;
       g = gtypes[mnum] ;
       if (qtmode) {  
         indx -> rawqval = genqval(qcc, indx -> thetatrue, a1/(a1+b1), a) ;
         indx -> qval = indx -> rawqval ;
       }
       if (!sim2dmode) break ;
       rlam = sim2dvals[2] ;
       rmu = sim2dvals[3]  ;
       kode = xlmrisk(a,g,rlam,rmu,numstates)  ;
       if (kode==YES)  break ;
       continue ;
      }
      if (indiv==1) {  
//     printf("zz %d %d %d %d %d %d\n", chrom, ethnic[mnum], gtypes[mnum], fc, lc, mnum) ;
      }
     }
     if (xnum1>=0) enum1 = ethnic[xnum1] ;
     if (xnum2>=0) enum2 = ethnic[xnum2] ;
/* support for epi */
     if ((sim2d_caseonly) && (markersim >= 0) && (indx -> affstatus == NO))
      gtypes[markersim] = -1 ;

       

     ivmaxmin(fchrom+1, 22, NULL, &fc) ;
     ivmaxmin(lchrom+1, 22, &lc, NULL) ;

     n1 = ntot = 0 ;
     for (k=fc ; k<= lc; ++k) {  
      cupt = snpmarkers[k] ;

      if (cupt -> isfake) continue ;
      if (cupt -> ignore) continue ;
      ntot += 2 ;
      n1 += ethnic[k] ;
     }
     ntot = MAX(ntot,1) ;
     y = (double) n1 / (double) ntot ;
     eauto[indiv] = y ;
     if (details && debug)
      printf("toy.  indiv %4d  %c %d ethnic auto: %9.3f %9.3f %9.3f",  
       indiv, indx -> gender, indx -> affstatus, 
       y, indx->theta_mode, indx->lambda_mode) ;
     fc = fchrom[23] ;  lc = lchrom[23] ;  
     if (fc>99999) noxdata = YES ;
     
     if (!noxdata) {
      cupt = snpmarkers[fc] ;
      gettln(cupt, indx, &theta, &lambda, &numstates, &ignore) ;
      numx = numstates - 1 ;
      n1 = ntot = 0 ;
      for (k=fc ; k<= lc; ++k) {  
       cupt = snpmarkers[k] ;
       if (cupt -> isfake) continue ;
       if (cupt -> ignore) continue ;
       ntot += numx ; 
       n1 += ethnic[k] ;
      }
      ntot = MAX(ntot,1) ;
      y = (double) n1 / (double) ntot ;
      ex[indiv] = y ;
      if (details &&debug)
       printf(" X: %9.3f thetaX  %9.3f\n",y, theta) ;
     }

// now copy gtypes into snpmarkers
     for (i=0; i<numsnps; i++)  {  
      cupt = snpmarkers[i] ;
      if (cupt -> isfake) continue ;
      if (cupt -> ngtypes == 0)  continue ;
      putgtypes(cupt, indiv, gtypes[i]) ;
     }
     if (indiv==1) {  
//    printf("zz %d %d %d\n", 999, ethnic[mnum], gtypes[mnum]) ;
     }
     
    } /* end indiv loop */

    y = asum(eauto, numindivs)/ (double) numindivs ;
    y2 = asum(tt, numindivs)/ (double) numindivs ;
    printf(" average ethnicity (autos) %9.3f\n",y) ;
    y = asum(ex, numindivs)/ (double) numindivs ;
    if (!noxdata) {
     y2 = asum(ttx, numindivs)/ (double) numindivs ;
     printf(" average ethnicity (X) %9.3f %9.3f\n",y, y2) ;
     y = corr(tt, ttx, numindivs) ;
     printf("correlation theory: %9.3f ", y) ;
     y = corr(eauto, ex, numindivs) ;
     printf("correlation actual: %9.3f", y) ;
    }
    printf("\n") ;

    free(ethnic2) ;
    free(ethnic) ;
    free(gtypes) ;
    free(eauto) ;
    free(ex) ;
    free(tt) ;
    free(ttx) ;

    printf("toysim exit\n") ;

}
void 
doepisim(SNP **snpm, Indiv **indm,
 int numsnps, int numindivs, double  *epirisk, double qrho,
 char *epinames, double *tauscal) 
{

  char *spt[2] ;  
  int  mnum1, mnum2 ;
  int n, indiv, a, b, k ; // k will be number of loci with at least on E-chromosome
  int niter = 0 ;
  Indiv *indx ;
  double t1[3], t2[9], t2r[9], theta, rr[9] ;
  double qq[9], qmax; 

  n = split1(epinames, spt, ':') ;
  if (n !=2 ) fatalx("bad epinames: %s\n", epinames) ; 
  xnum1 = mnum1 = snpindex(snpm, numsnps, spt[0]) ;
  xnum2 = mnum2 = snpindex(snpm, numsnps, spt[1]) ;
  printf("doepisim: %d %d\n", mnum1, mnum2) ;
  if ( (mnum1<0) || (mnum2<0)) fatalx("bad epi indices\n") ;
  freeup(spt, 2) ;

  for (a=0; a<3; ++a) {  
   for (b=0; b<3; ++b) {  
    k = 0; 
    if (a>=1) ++k ;  
    if (b>=1) ++k ;  
    rr[3*a+b] = epirisk[k] ;
   }
  }

 for (indiv=0; indiv<numindivs; ++indiv) { 
 indx  = indm[indiv] ;
 if (indx -> ignore) continue ;
  for (;;) {
   ++niter ;
  if (niter >= 1000000) fatalx("(episim) looping\n") ;
   dosim(snpm, indm,
    numsnps, numindivs, 1.0, qrho,
    -1, tauscal) ;
    if (indx -> affstatus == NO)  break ;
    theta = indx -> theta_mode ;
    sett1(t1, theta, 3) ; // autosomes only
    vzero(t2, 9) ; 
    addouter(t2, t1, 3) ;
    vvt(t2r, t2, rr, 9) ; 
    bal1(t2, 9) ; 
    bal1(t2r, 9) ;
    vvd(qq, t2r, t2, 9) ;
    vmaxmin(qq, 9, &qmax, NULL) ; /* max ratio */
    vst(qq, qq, 1.0/qmax, 9) ;
    a = enum1 ;  b = enum2 ;
    if (DRAND() < qq[3*a+b]) break ; /* got it */
   }
 }
}

int xlmrisk(int a, int g, double lam, double mu, int numstates) 
// rejection sampler 
// reject (a,g) pair according to 2d risk model
{
 int gg ;
 double maxlam, maxmu, prob ;
 if (a<0) return YES ;
 if (g<0) return YES ;
 gg = g ;
 if (numstates==2) gg = g/2 ;
 if (lam<1.0) maxlam = 1.0 ;        
  else maxlam = pow(lam, (double) numstates) ;
 if (mu<1.0) maxmu = 1.0 ;        
  else maxmu = pow(mu, (double) numstates) ;
 prob = pow(lam,(double) a) * pow(mu, (double) gg) ;
 prob /= (maxlam*maxmu) ;
 if (DRAND() < prob) return YES ;
 return NO ;
}

void gx (double *pp, double p, double tau) 
{
   *pp = ranbeta(p*tau, (1.0-p)*tau) ;
}

void 
ranpathx(double theta, double lambda,  int numstates,
 SNP **snpmarkers, int *ethnic, 
 int fc, int lc) 
/* like ranpath but no betas */

{

 int len, count ;
 SNP *cupt ;
 Indiv *indx ;
 double bb[3] , cc[3],  y ;
 int i, j, k, l; 
 double **M ;
 double lfac[3],  logfac ;
 double glen, poisson, temp ;
 int estate, oldestate ;
 int debugg = NO ;

 cupt = snpmarkers[fc] ;

  count = len = lc -fc +1 ;
  ivclear(ethnic+fc, -1, len) ;
  M = initarray_2Ddouble(3, 3, 0.0);


  sett1(bb, theta, numstates) ;
  estate = randis(bb, numstates) ;
/* actually estate is irrrelevant but this is careful */

  logfac = 0.0 ;

  for (j=fc; j <= lc; ++j) {  

    count = j-fc ;
    oldestate = estate ;
    poisson = 0.0 ;
    if (j>fc) {
     glen = (snpmarkers[j] -> genpos) - (snpmarkers[j-1] -> genpos);
/* this is in morgans */ 
     temp = lambda * glen ;
     if (temp<20.0) {
      poisson = exp(-temp) ;
     }
     if (lambda>500) poisson = 0.0 ;
    }

    calcM(poisson, theta, M, numstates);

     for (k=0; k<numstates; k++)  {  
      cc[k] = M[k][oldestate] ;
     }

     y = asum(cc, numstates) ;
     vst(bb, cc, 1.0/y, numstates) ;
     ethnic[j] = estate = randis(bb, numstates) ;
   }
  free2D(&M, 3) ;
}

void 
ranpathr(double theta, double lambda,  int numstates,
 SNP **snpmarkers, int *ethnic, 
 int fc, int lc, int markersim, double risk) 
{
   double tt[3], ttr[3], qq[3] ;
   double qmax, xrisk ;
   int a ;
   int iter = 0, iterlim ;

   xrisk = risk ; 
   if (markersim<0) xrisk = 1.0 ;
   sett1(tt, theta, numstates) ;
   sett1r(ttr, theta, numstates, xrisk) ;

   if ((theta==1.0)  || (theta==0.0) || (lambda==0.0)) {
    a = randis(ttr, numstates) ;
    ivclear(ethnic+fc, a, lc-fc+1) ;
    return ;
   }


   vvd(qq, ttr, tt, numstates) ;
   vmaxmin(qq, numstates, &qmax, NULL) ; /* max ratio */
   vst(qq, qq, 1.0/qmax, numstates) ;

   iterlim = nnint(100.0*qmax) ;

/** now do rejection sampling 
  We should be drawing ethnic[] with dist ttr and 
  ranpathx samples with distribution tt 
*/
   for (;;) {  
     ranpathx(theta, lambda, numstates, snpmarkers, ethnic, 
      fc, lc) ;
     if ((markersim<fc) || (markersim>lc)) return ;
     if (xrisk == 1.0) return ;
     a = ethnic[markersim] ;
     if (a<0) fatalx("(ranpathr) bad ethnic %d %d\n",markersim, a) ;
     if (DRAND() <= qq[a]) return ;   /* accept */
     /* try again */
     ++iter ;
     if (iter >= iterlim) fatalx("(ranpathr) ... looping\n") ;
   }
}

void
mkgtypes(int *gtypes, SNP **tsnpm, int *ethnic, int indiv, int numstates, 
 int fc, int lc)  
{

  int i, n, a, xx[2], gg[2] ;  
  int j, numx  ; 
  double p, y ; 
  SNP *cupt ;

  static int ncall = 0 ; 
  static int nmiss ; 

  ++ncall ;
  numx = numstates -1 ;
  for (i=fc ; i<= lc; i++)  { 
   cupt = tsnpm[i] ;
   if (cupt -> isfake) continue ;
   if (cupt -> ignore) continue ;
   n = cupt -> ngtypes ;

   if (indiv >= n) continue ;
   a = getgtypes(cupt, indiv)  ;  
   if ((a<0) && (!alldata))  {
    gtypes[i] = a ;  
//  if (nmiss < 10) printf("zzmiss: %d %d %d\n", i, indiv, a) ; 
    ++nmiss ; 
    continue ;
   }
   a = ethnic[i] ;
   genaa(a, xx, numstates) ;
   gtypes[i] = 0 ;
   for (j=0; j<numx ; ++j)  {  
    y = DRAND() ;
    if (xx[j] == 0) p = cupt -> adm_PopA_freq ;
    if (xx[j] == 1) p = cupt -> adm_PopB_freq ;
    if (y < p)  ++gtypes[i] ;  
   }
   if (numstates==2) gtypes[i] *= 2 ;
  }
} 

double ranx(double z)  
{
  double t ;
  t = 0.05*gauss() ;
  t = MIN(t, 0.1) ;
  t = MAX(t, -0.1) ;
  return (z+t) ;
}

double genth1(double a, double b) 

{
/** 
 idea is pick theta ranbeta(a,b) with prob 1-alpha 
 else pick 0:1 with p(1) = a/a+b
 such that mean variance are same as ranbeta(a,b) 
*/

    double m, p2, top, bot, alpha ;
    static int ncall = 0 ;

    ++ncall ;

    m = a/(a+b) ;

    top = a*b ;
    bot = (a+b)*(a+b)*(a+b+1) ;
    p2 = top/bot ;

    top = p2 ; 
    bot = (m-m*m) - p2 ;
    alpha = top/bot ;

    if (ncall==1) printf("zz genth1 (alpha): %9.3f\n",alpha) ;

    if (DRAND() < alpha) {  
     if (DRAND() < m) return 1.0 ;
     return 0.0 ;
    }
    return ranbeta(a,b) ;
}

double
genlam1(double p, double psi, int mode) 
{

 double m,v,m2,mx,ac,alpha,vx,p2,l2  ;
 static int ncall = 0 ;

 ++ncall ;
 if (mode==0) return rangam(p)/psi ;

 m = p/psi ;
 v = m/psi ;
 m2 = m*m + 2*v ;
/* we need 2*v since we are doing this for single chromosome
   not 2 
*/


 mx = m2/m - 1 ;
 ac = m/(mx+1) ;
 alpha = 1-ac ;
 
 if (mode == 1) {
  if (ncall==1) 
   printf("mx: %9.3f alpha: %9.3f\n", mx, alpha) ;
  if (DRAND() < alpha) return 0 ;
  return mx+1 ;
 }

 mx = m -1  ;
 vx = m2 - m*m ;

 l2 = mx/vx ; /* lambda */
 p2 = mx*l2 ;

 if (ncall==1) 
  printf("mx: %9.3f vx: %9.3f p2: %9.3f l2: %9.3f\n", 
   mx, vx, p2, l2) ;

 return (1.0+rangam(p2)/l2) ;

}

void genindiv(Indiv **indivmarkers, int ncases, int ncontrols, 
 int mode, int allmale) 
{
/** 
 mode 1 
 ncases ncontrols match 
 in blocks of 4 just set 2 cases 2 controls 2 males 2 females 

 mode 2 
 first chunk of individuals are cases 

 This is so as to make it easy to nest problems 
*/

  int numindivs, i, k ;
  Indiv *indx ;
  
  numindivs = ncases + ncontrols ;

  if (mode == 1) {
   for (i=0; i<numindivs; i++) {  
    indx = indivmarkers[i] ;
    indx -> ignore = NO ;
    indx -> affstatus = YES ;
    indx -> gender = 'F' ;
    k = i%4 ;
    if (k<2) indx -> affstatus = NO ;
    k = i%2 ;
    if (k == 0) indx -> gender = 'M'   ;
    if (allmale) indx -> gender = 'M'  ;
   }
  }
   for (i=0; i<numindivs; i++) {  
    if (mode == 1) break ;
    indx = indivmarkers[i] ;
    indx -> ignore = NO ;
    indx -> affstatus = YES ;
    indx -> gender = 'F' ;
    if (i >= ncases) indx -> affstatus = NO ;
    k = i%2 ;
    if (k == 0) indx -> gender = 'M'   ;
    if (allmale) indx -> gender = 'M'  ;
   }
   for (i=0; i<numindivs; i++) {  
    indx = indivmarkers[i] ;
    if (indx -> affstatus) indx -> egroup = strdup("Case") ;
    else indx -> egroup = strdup("Control") ;
   


   }
}
